/**
 *
 * @file 8bit_bootloader.c
 *
 * @ingroup generic_bootloader_8bit
 *
 * @brief This source file provides the implementation of the APIs for the 8-bit Bootloader library
 *
 * @version BOOTLOADER Driver Version 3.0.0
 */

${disclaimer}

<#if CHIP_TYPE == "PIC16">
//   Memory Map
//   -----------------
//   |    0x0000     |   Reset vector
//   |               |
//   |    0x0004     |   Interrupt vector
//   |               |
//   |               |
//   |  Boot Block   |   (this program)
//   |               |
//   |    ${NEW_RESET_VECTOR_U}     |   Re-mapped Reset Vector
//   |    ${NEW_INTERRUPT_VECTOR_LOW_U}     |   Re-mapped High Priority Interrupt Vector
//   |               |
//   |       |       |
//   |               |
//   |  Code Space   |   User program space
//   |               |
//   |       |       |
//   |               |
//   |    0x3FFF     |
//   -----------------
//
</#if>
<#if CHIP_TYPE == "PIC18">
//   Memory Map
//   -----------------
//   |    0x0000     |   Reset vector
//   |               |
//   |    0x0008     |   High Priority Interrupt vector
//   |               |
//   |    0x0018     |   Low Priority Interrupt vector
//   |               |
//   |  Boot Block   |   (this program)
//   |               |
//   |    ${NEW_RESET_VECTOR_U}     |   Re-mapped Reset Vector (Actual address
//   |    ${NEW_INTERRUPT_VECTOR_HIGH_U}     |   Re-mapped High Priority Interrupt Vector
//   |    ${NEW_INTERRUPT_VECTOR_LOW_U}     |   Re-mapped Low Priority Interrupt Vector
//   |       |       |
//   |               |
//   |  Code Space   |   User program space
//   |               |
//   |       |       |
//   |               |
//   | End of Flash  |
//   -----------------
//
</#if>
<#if DEVICE_TYPE == "AVR">
//   Memory Map
//   ----------------
//   |   0x000000   |   PROGMEM_START
//   |              |
//   |      |       |
//   |     BOOT     |  (this program)
//   |      |       |
//   |              |
//   |   Boot End   |   FUSE_BOOTSIZE * PROGMEM_PAGE_SIZE
//   |              |
//   |      |       |
//   |     APP      |   (User program space)
//   |      |       |
//   |              |
//   |   App End    |   FUSE_CODESIZE * PROGMEM_PAGE_SIZE
//   |              |
//   |      |       |
//   |   APPDATA    |   App Data Space
//   |      |       |
//   |              |
//   |   ${FLASH_SIZE_U}   |   PROGMEM_END
//   ----------------
//
</#if>
// *****************************************************************************

#include <stdbool.h>
#include "../bl_bootload.h"
#include "../bl_communication_interface.h"
<#if DEVICE_TYPE == "AVR">
#include <stdlib.h>
#include "../../${DELAY_HEADER_PATH}"
#include <string.h>
</#if>

//****************************************
// Default Functions (Always Used)
static uint8_t BL_GetVersionData(void);
static void BL_RunBootloader(void);
static bool BL_BootloadRequired(void);
static void BL_CheckDeviceReset(void);
static uint8_t BL_WriteFlash(void);
static uint8_t BL_EraseFlash(void);
static uint16_t BL_ProcessBootBuffer(void);

//****************************************
// Conditional Functions
<#if SUPPORT_CHECKSUM == "Enabled">
static uint8_t BL_CalcChecksum(void);
</#if>
<#if SUPPORT_FLASH_READ == "Enabled">
static uint16_t BL_ReadFlash(void);
</#if>
<#if SUPPORT_CONFIG_READ == "Enabled">
static uint16_t BL_ReadConfig(void);
</#if>
<#if SUPPORT_CONFIG_WRITE == "Enabled">
static uint8_t BL_WriteConfig(void);
</#if>
<#if VERIFY == "STATUS_BYTE_VERIFICATION">
static uint8_t BL_WriteAppStatus(void);
</#if>
<#if SUPPORT_EE_DATA_READ == "Enabled">
static uint16_t BL_ReadEEData(void);
</#if>
<#if SUPPORT_EE_DATA_WRITE == "Enabled">
static uint8_t BL_WriteEEData(void);
</#if>


<#if DEVICE_TYPE == "AVR">
typedef void(*app_t)(void);
static uint8_t BL_Reset(void);
</#if>

<#if CHIP_TYPE == "PIC16">
// *****************************************************************************
// The bootloader code does not use any interrupts.
// However, the application code may use interrupts.
// The interrupt vector on a PIC16F is located at
// address 0x0004. 
// The following function will be located
// at the interrupt vector and will contain a jump to
// the new application interrupt vector
asm ("psect  intentry,global,class=CODE,delta=2");
asm ("pagesel " str(NEW_INTERRUPT_VECTOR));
asm ("GOTO " str(NEW_INTERRUPT_VECTOR));
</#if>

<#if CHIP_TYPE == "PIC18">
<#if (( DEVICE_FAMILY == "K42") || ( DEVICE_FAMILY == "K83") || (PIC18F_Q == "Enabled"))>
// *****************************************************************************
// Newer PIC18F devices have vectored interrupts with the base address specified
// by SFR IVTBASE.  Thus there is no need for the bootloader to redirect the
// Interrupt vectors. The end application should specify IVTBASE as needed
// *****************************************************************************
<#else>
// *****************************************************************************
// The bootloader code does not use any interrupts.
// However, the application code may use interrupts.
// The interrupt vector on a PIC18F is located at
// address 0x0008 (High) and 0x0018 (Low). 
// The following function will be located
// at the interrupt vector and will contain a jump to
// the new application interrupt vectors
// *****************************************************************************
asm ("psect  intcode,global,reloc=2,class=CODE,delta=1");
asm ("GOTO " str(NEW_INTERRUPT_VECTOR_HIGH));

asm ("psect  intcodelo,global,reloc=2,class=CODE,delta=1");
asm ("GOTO " str(NEW_INTERRUPT_VECTOR_LOW));
</#if>
</#if> 
// *****************************************************************************
static bool resetPending = false;

// The data frame used for
// holding the current data frame throughout
// boot operation
static frame_t frame;

/*
 * @todo Documentation Needed
 */
void BL_Initialize(void)
{
    resetPending = false;

<#if (BOOTLOADER_RUNNING_IO == "Enabled")>
    BL_INDICATOR_OFF();
</#if>

    if (BL_BootloadRequired() == true)
    {
<#if DEVICE_TYPE == "AVR">
        DELAY_milliseconds(200U);
</#if>

<#if (BOOTLOADER_RUNNING_IO == "Enabled")>
        BL_INDICATOR_ON();
</#if>
        BL_RunBootloader(); // generic comms layer
    }
<#if (CHIP_TYPE == "PIC18")>
    STKPTR = 0x00U;
    BSR = 0x00U;
</#if>
<#if CHIP_TYPE == "PIC16">
    STKPTR = 0x1FU;
    BSR = 0U;
</#if>
<#if (BOOTLOADER_RUNNING_IO == "Enabled")>
    BL_INDICATOR_OFF();
</#if>
<#if DEVICE_TYPE == "PIC">
<#if CHIP_TYPE == "PIC16">
    asm("pagesel " str(NEW_RESET_VECTOR));
</#if>
    asm("goto  " str(NEW_RESET_VECTOR));
</#if>
<#if DEVICE_TYPE == "AVR">
    app_t app = (app_t)(START_OF_APP / sizeof(app_t));
    app();    
</#if>
}

static bool BL_BootloadRequired(void)
{
    bool status;
<#if IO_PIN_ENTRY == "Enabled">
    /* NOTE: This function can be customized by the user to enter the bootloader 
     * as required. The entry point can be reading a push button status, a  
     * command from a peripheral, or any other method selected by the user.  
     * 
     * Currently the bootloader checks an IO pin at programming time to force entry into bootloader.
    */  

    // #info  "You may need to add additional delay here between enabling weak pullups and testing the pin."
    for (uint8_t i = 0U; i != 0xFFU; i++)
    {
<#if DEVICE_TYPE == "AVR">
        asm("nop");
<#else>
        NOP();
</#if>
    }
    if (IO_PIN_ENTRY_GetInputValue() == IO_PIN_ENTRY_RUN_BL)
    {
            return (true);                
    }
</#if>
    if (BL_bootVerify() == false)
    {
        status = true;
    }
    else
    {
        status = false;
    }

    return status;
}

/**
 * @ingroup generic_bootloader_8bit
 * @brief Processes the command header and returns the length of the return packet.
 * @param none
 * @retval The total length of the packet being passed back to the host.
 */
static uint16_t BL_ProcessBootBuffer(void)
{
    uint16_t len;
    switch (frame.command)
    {
    case READ_VERSION:
        len = BL_GetVersionData();
        break;
<#if (SUPPORT_FLASH_READ == "Enabled")>
    case READ_FLASH:
        len = BL_ReadFlash();
        break;
</#if>
    case WRITE_FLASH:
        len = BL_WriteFlash();
        break;
    case ERASE_FLASH:
        len = BL_EraseFlash();
        break;
<#if (SUPPORT_EE_DATA_READ == "Enabled")>
    case READ_EE_DATA:
        len = BL_ReadEEData();
        break;
</#if>
<#if (SUPPORT_EE_DATA_WRITE == "Enabled")>
    case WRITE_EE_DATA:
        len = BL_WriteEEData();
        break;
</#if>
<#if SUPPORT_CONFIG_READ == "Enabled">    
    case READ_CONFIG:
        len = BL_ReadConfig();
        break;
</#if>
<#if SUPPORT_CONFIG_WRITE == "Enabled">        
    case WRITE_CONFIG:
        len = BL_WriteConfig();
        break;
</#if>
<#if SUPPORT_CHECKSUM == "Enabled">
<#if CHIP_TYPE == "PIC18" || CHIP_TYPE == "PIC16">
    case CALC_CHECKSUM:
        len = BL_CalcChecksum();
        break;
<#else>
    case CALC_CHECKSUM:
        // not supported yet
        frame.data[0] = ERROR_INVALID_COMMAND;
        len = BL_HEADER + 1U;
        break;
</#if>
</#if>
    case RESET_DEVICE:
<#if DEVICE_TYPE == "AVR">
        frame.data[0] = COMMAND_SUCCESS;
        len = BL_Reset();
<#else>
            frame.data[0] = COMMAND_SUCCESS;
        resetPending = true;
        len = 10U;
</#if>
        break;
<#if VERIFY == "STATUS_BYTE_VERIFICATION">
    case WRITE_STATUS:
        len = BL_WriteAppStatus();
        break;
</#if>
    default:
        frame.data[0] = ERROR_INVALID_COMMAND;
        len = 10U;
        break;
    }
    return (len);
}

static void BL_RunBootloader(void)
{
    uint16_t messageLength = 0U;
    uint16_t index = 0U;
    uint8_t ch;

    while (1)
    {
        BL_CheckDeviceReset();

        BL_CommunicationModuleInit();

        index = 0U; //Point to the buffer
        messageLength = BL_HEADER; // message has 9 bytes of overhead (Synch + Opcode + Length + Address)
<#if DEVICE_TYPE == "AVR">
        memset(frame.buffer,0,sizeof(frame));    // Clear Buffer
</#if>
        while (index < messageLength)
        {
            BL_CommunicationModuleRead(&ch, 1);
            frame.buffer[index] = ch;

            index++;
            if (index == 5U)
            {
                if ((frame.command == WRITE_FLASH)
                        || (frame.command == WRITE_EE_DATA)
                        || (frame.command == WRITE_CONFIG))
                {
                    messageLength += frame.data_length;
                }
<#if VERIFY == "STATUS_BYTE_VERIFICATION">
                else if (frame.command == WRITE_STATUS)
                {
                    messageLength += 1U;
                }
</#if>
                else
                {
                    //do nothing
                }
            }
        }

        messageLength = BL_ProcessBootBuffer();

        if (messageLength > 0U)
        {
            BL_CommunicationModuleWrite(frame.buffer, messageLength);

            while (BL_CommunicationModuleIsReady() != true)
            {

            }
        }
    }
}

static void BL_CheckDeviceReset(void)
{
    if (resetPending == true)
    {
<#if (DEVICE_TYPE == "AVR")>
<#if (VERIFY == "STATUS_BYTE_VERIFICATION")>
<#if (NVM_ACCESS_TYPE == "BYTE_ACCESS")>
        nvm_status_t status = NVM_OK;
        if(FLASH_Read(STATUS_ADDRESS) == (flash_data_t) 0xFFU)
        {
            status = FLASH_Write(STATUS_ADDRESS, (uint16_t)APPLICATION_VALID);
        }
</#if>
<#if (NVM_ACCESS_TYPE == "ROW_ACCESS")>
        flash_address_t flashStartPageAddress;
        uint16_t flashAddressOffset;
        flash_data_t flashWriteData[PROGMEM_PAGE_SIZE];

        // Read out the page that holds the status byte
        if(FLASH_Read(STATUS_ADDRESS) == (flash_data_t) 0xFFU){
            //Get the starting address of the page containing given address
            flashStartPageAddress = FLASH_PageAddressGet(STATUS_ADDRESS);

            //Read entire row
            for (flashAddressOffset = 0U; flashAddressOffset < PROGMEM_PAGE_SIZE; flashAddressOffset++)
            {
                flashWriteData[flashAddressOffset] = FLASH_Read(flashStartPageAddress + flashAddressOffset);
            }

            //Get offset from starting address of the page
            flashAddressOffset = FLASH_PageOffsetGet(STATUS_ADDRESS);

            // Set the status value
            flashWriteData[flashAddressOffset] = APPLICATION_VALID;

            //Write data to Flash row
            FLASH_RowWrite(flashStartPageAddress, flashWriteData);
        }
</#if>
</#if>
        DELAY_milliseconds(3U);
            ccp_write_io((void *)&RSTCTRL.SWRR, ${RESET_BIT_U});
    }
</#if>
<#if CHIP_TYPE == "PIC18">
<#if (VERIFY == "STATUS_BYTE_VERIFICATION")>
        // *************************************
        // Write the status of the bootloader at the device reset
        // which should occur after every reprogramming action
        // *************************************
        if (FLASH_Read((flash_address_t) STATUS_ADDRESS) == (flash_data_t) 0xFFU)
        {
            uint16_t unlockKey;
            nvm_status_t updateStatus;

            unlockKey = (((uint16_t) frame.EE_key_2) << 8U)
                    | (uint16_t) frame.EE_key_1;

            NVM_StatusClear();

            NVM_UnlockKeySet(unlockKey);

            updateStatus = FLASH_Write((flash_address_t) STATUS_ADDRESS, (uint16_t) APPLICATION_VALID);

            NVM_UnlockKeyClear();

            if (updateStatus == NVM_ERROR)
            {
                NVM_StatusClear();
            }
        }
</#if>
<#if (BOOTLOADER_RUNNING_IO == "Enabled")>
        BL_INDICATOR_OFF();
</#if>
        RESET();
    }
    return;
</#if>
<#if CHIP_TYPE == "PIC16">
<#if VERIFY == "STATUS_BYTE_VERIFICATION">
        // *************************************
        // Write the status of the bootloader at the device reset
        // which should occur after every reprogramming action
        // *************************************
        if( FLASH_Read(STATUS_ADDRESS) == (flash_data_t) 0x3FFFU)
        {   
            flash_address_t flashStartPageAddress;
            nvm_status_t status = NVM_OK;
            flash_data_t wordBuff[PROGMEM_PAGE_SIZE];
            uint16_t unlockKey = (((uint16_t)frame.EE_key_2) << 8U)
                                 | (uint16_t)frame.EE_key_1;
            
            flashStartPageAddress = FLASH_PageAddressGet(STATUS_ADDRESS);
            
            for (uint8_t i = 0U; i < PROGMEM_PAGE_SIZE; i++)
            {
                wordBuff[i] = FLASH_Read(flashStartPageAddress + i);
            }
            wordBuff[PROGMEM_PAGE_SIZE - 1U] = (uint16_t)APPLICATION_VALID;
            
            NVM_UnlockKeySet(unlockKey);
            status = FLASH_RowWrite(flashStartPageAddress, wordBuff);
            NVM_UnlockKeyClear();

            if (status == NVM_ERROR)
            {
                NVM_StatusClear();
            }
        }
</#if>

<#if (BOOTLOADER_RUNNING_IO == "Enabled")>
        BL_INDICATOR_OFF();
</#if>
        RESET();
    }
</#if>
}

// ******************************************************************************
// Get Bootloader Version Information
//        Cmd     Length----------------   Address---------------
// In:   [|0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00|]
// OUT:  [|0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | VERL | VERH|]
// ******************************************************************************

static uint8_t BL_GetVersionData(void)
{
    uint8_t dataIndex = 0U;
    uint32_t maxPacketSize = 0U;

    maxPacketSize = (PROGMEM_SIZE / ((uint32_t) PROGMEM_PAGE_SIZE));
<#if CHIP_TYPE == "PIC16">
    // 0x06 is the device id address offset for NVMREG access on newer PIC16F devices
    device_id_data_t deviceId = DeviceID_Read(DEVICE_ID_START_ADDRESS);
<#elseif CHIP_TYPE == "PIC18">
    device_id_data_t deviceId = DeviceID_Read(DEVICE_ID_START_ADDRESS);
<#elseif DEVICE_TYPE == "AVR">
    device_id_data_t deviceId = SIGROW_DeviceIDRead(SIGNATURES_START);
</#if>

    // Bootloader Firmware Version
    frame.data[dataIndex] = MINOR_VERSION;
    dataIndex++;
    frame.data[dataIndex] = MAJOR_VERSION;
    dataIndex++;
<#if DEVICE_TYPE == "AVR">
    frame.data[dataIndex] = BL_HEADER;
    dataIndex++;

    frame.data[dataIndex] = SIGROW_Read(FUSES_START + 0x08U); // BOOT FUSE
    dataIndex++;
    frame.data[dataIndex] = SIGROW_Read(FUSES_START + 0x07U); // APP FUSE
    dataIndex++;
</#if>

    // max packet size in hexadecimal
    frame.data[dataIndex] = (uint8_t) (maxPacketSize & 0xFFU);
    dataIndex++;
    frame.data[dataIndex] = (uint8_t) ((maxPacketSize >> 8U) & 0xFFU);
    dataIndex++;

    // Unused Bytes
    frame.data[dataIndex] = 0U;
    dataIndex++;
    frame.data[dataIndex] = 0U;
    dataIndex++;

    // device id
    frame.data[dataIndex] = (uint8_t) deviceId;
    dataIndex++;
    frame.data[dataIndex] = (uint8_t) (deviceId >> 8U);
    dataIndex++;
<#if DEVICE_TYPE == "AVR">
    frame.data[dataIndex] = (uint8_t)(deviceId >> 16U);
    dataIndex++;
<#else>
    // Unused Bytes
    frame.data[dataIndex] = 0U;
    dataIndex++;
    frame.data[dataIndex] = 0U;
    dataIndex++;
</#if>
    frame.data[dataIndex] = (uint8_t) (PROGMEM_PAGE_SIZE & 0xFFU);
    dataIndex++;
    frame.data[dataIndex] = (uint8_t) (((uint16_t)PROGMEM_PAGE_SIZE >> 8U) & 0xFFU);
    dataIndex++;

<#if DEVICE_TYPE == "PIC">
    // Read 4 bytes of the user id
<#if CHIP_TYPE == "PIC16">
    uint16_t offsetAddress = 0U;
<#else>
    uint24_t offsetAddress = USER_ID_START_U;
</#if>
    for (uint8_t i = 0U; i < 4U; i++)
    {
<#if (CHIP_TYPE == "PIC18")>
        frame.data[dataIndex] = (uint8_t) FLASH_Read(offsetAddress);
<#else>
        frame.data[dataIndex] = (uint8_t) CONFIGURATION_Read(offsetAddress);
</#if>     
        offsetAddress++;
        dataIndex++;
    }
</#if>
<#if DEVICE_TYPE == "AVR">
    frame.data[dataIndex] = SIGROW.SERNUM0;
    dataIndex++;
    frame.data[dataIndex] = SIGROW.SERNUM1;
    dataIndex++;
    frame.data[dataIndex] = SIGROW.SERNUM2;
    dataIndex++;
    frame.data[dataIndex] = SIGROW.SERNUM3;
    dataIndex++;
    frame.data[dataIndex] = SIGROW.SERNUM4;
    dataIndex++;
    frame.data[dataIndex] = SIGROW.SERNUM5;
    dataIndex++;
    frame.data[dataIndex] = SIGROW.SERNUM6;
    dataIndex++;
    frame.data[dataIndex] = SIGROW.SERNUM7;
    dataIndex++;
    frame.data[dataIndex] = SIGROW.SERNUM8;
    dataIndex++;
    frame.data[dataIndex] = SIGROW.SERNUM9;
    dataIndex++;
</#if>

    return (BL_HEADER + dataIndex); // total length to send back 9 byte header + payload
}

<#if (SUPPORT_FLASH_READ == "Enabled")>
<#if DEVICE_TYPE == "AVR">
// *********************************************************************************************************************
// Request Flash Read over range using Bootloader
// [|CMD | Data Length(Low, High) Bytes to Read | unused | unused | Address(Low,High,Upper,Extended)to Start Reading From|]
// IN: [|0x01 | DataLengthL | DataLengthH | unused | unused | ADDRL | ADDRH | ADDRU | ADDRE>...]
// OUT: [|0x01 | DataLengthL | DataLengthH | unused | unused | ADDRL | ADDRH | ADDRU | ADDRE | DATA>...]
// *********************************************************************************************************************
</#if>
<#if CHIP_TYPE == "PIC18">
// *****************************************************************************
// Read Flash
//        Cmd     Length----------------   Address---------------  Data ---------
// In:   [|0x01 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00|]
// OUT:  [|0x01 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | Data |.. | data |]
// *****************************************************************************
</#if>
<#if CHIP_TYPE == "PIC16">
// **************************************************
// In:    [|0x01 | DLEN | ADDRL | ADDRH | ADDRU|]
// OUT:   [|0x01 | DLEN | ADDRL | ADDRH | ADDRU | DATA>...]
// **************************************************
</#if>
static uint16_t BL_ReadFlash(void)
{
    flash_address_t address;
    uint16_t dataIndex;
<#if CHIP_TYPE == "PIC16">
    flash_data_t wordRead;
</#if>

<#if CHIP_TYPE == "PIC16" || (DEVICE_TYPE == "AVR" && AVR_PAGE == "Classic")>
    address = (((flash_address_t)frame.address_H) << 8U) 
            | (flash_address_t)frame.address_L;
<#elseif (DEVICE_TYPE == "AVR" && AVR_PAGE == "Expanded")>
    address = ( ( (flash_address_t)frame.address_E << 24U)
                    | ( (flash_address_t)frame.address_U << 16U)
                    | ( (flash_address_t)frame.address_H << 8U)
                    | frame.address_L);
<#elseif CHIP_TYPE == "PIC18">
    address = (((flash_address_t) frame.address_U) << 16U)
            | (((flash_address_t) frame.address_H) << 8U)
            | (flash_address_t) frame.address_L;
</#if>

    /*
     * Note: 
     *      This block is used to prevent read access within the boot block and outside of the normal flash range. 
     *      This was done because it is considered a security risk to allow the host to read the code out of the bootloader. 
     *   
     *      If you want to allow this type of read access in your system, you will need to delete this if block. And be warned that 
     *      a malicious host could read portions of the chips memory that should be considered private.
     */
    if(!((address >= START_OF_APP) && (address < PROGMEM_SIZE)))
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10U);
    }

    // Prevent any read operation that exceeds the data buffer size
    if( frame.data_length > BL_FRAME_DATA_SIZE )
    {
        frame.data[0] = COMMAND_OVERLOAD_ERROR;
        return (10U);
    }

<#if DEVICE_TYPE == "AVR" || CHIP_TYPE == "PIC18">
    for (dataIndex = 0U; dataIndex < frame.data_length; dataIndex++)
    {
        frame.data[dataIndex + 1U] = FLASH_Read(address);
        ++address;
    }
    frame.data[0] = COMMAND_SUCCESS;
</#if>
<#if CHIP_TYPE == "PIC16">
    for(dataIndex = 0; dataIndex < frame.data_length; dataIndex += 2U)
    {
        wordRead = FLASH_Read(address);
        frame.data[dataIndex + 1U] = (uint8_t)(wordRead & 0x00FFU);
        frame.data[dataIndex + 2U] = (uint8_t) ((wordRead & 0xFF00U) >> 8U);
        ++address;
    }
    frame.data[0] = COMMAND_SUCCESS;
</#if>

    return (frame.data_length + 10U);
}
</#if>

<#if DEVICE_TYPE == "AVR">
// ********************************************************************************************************************
// Request Flash Write over range, with supplied data using Bootloader
// [|CMD | Data Length(Low, High) Bytes to Write | unused | unused | Address(Low,High,Upper,Extended)to Start Writing To|]
// IN: [|0x02 | DataLengthL | DataLengthH | unused | unused | ADDRL | ADDRH | ADDRU | ADDRE | DATA>...]
// OUT: [|0x02 | DataLengthL | DataLengthH | unused | unused | ADDRL | ADDRH | ADDRU | ADDRE | STATUS_RESPONSE>...]
// ********************************************************************************************************************
</#if>
<#if DEVICE_TYPE == "PIC">
// *****************************************************************************
// Write Flash
//        Cmd     Length----- Keys------   Address---------------  Data ---------
// In:   [|0x02 | 0x00 | 0x00 | 0x55 | 0xAA | 0x00 | 0x00 | 0x00 | 0x00 | Data |.. | data |]
// OUT:  [|0x02 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x01|]
// *****************************************************************************
</#if>
static uint8_t BL_WriteFlash(void)
{
    nvm_status_t errorStatus = NVM_OK;
    flash_address_t userAddress;
    flash_address_t flashStartPageAddress;
    flash_address_t userDataStartOffset;
    flash_data_t writeBuffer[PROGMEM_PAGE_SIZE];

<#if DEVICE_TYPE == "PIC">
    uint16_t unlockKey = (((uint16_t) frame.EE_key_2) << 8U) 
                        | (uint16_t) frame.EE_key_1;

    if(unlockKey != UNLOCK_KEY)
    {
        frame.data[0] = COMMAND_PROCESSING_ERROR;
        return (10U);
    }

    <#if CHIP_TYPE == "PIC16">
    userAddress = (((flash_address_t)frame.address_H) << 8U)
        |   (flash_address_t)frame.address_L;
    
    </#if>
    <#if CHIP_TYPE == "PIC18">
    userAddress = (((flash_address_t) frame.address_U) << 16U)
            | (((flash_address_t) frame.address_H) << 8U)
            | (flash_address_t) frame.address_L;
    </#if>
</#if>
<#if DEVICE_TYPE == "AVR">
    <#if (AVR_PAGE == "Expanded")>
    userAddress = ( ( (flash_address_t)frame.address_E << 24U) 
                    | ( (flash_address_t)frame.address_U << 16U) 
                    | ( (flash_address_t)frame.address_H << 8U) 
                    | frame.address_L);
    </#if> 
    <#if (AVR_PAGE == "Classic")>
    userAddress = ( ( (flash_address_t)frame.address_H << 8U) 
                    | frame.address_L);
    </#if>
    
    if (userAddress < START_OF_APP)
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10U);
    }
</#if>

    // Prevent any write operation that exceeds the data buffer size
    if( frame.data_length > BL_FRAME_DATA_SIZE )
    {
        frame.data[0] = COMMAND_OVERLOAD_ERROR;
        return (10U);
    }

<#if (DEVICE_TYPE == "PIC" && SELF_PROTECTION == "Enabled")>
    <#if CHIP_TYPE == "PIC16">
    if ((userAddress & 0x7FFF) < NEW_RESET_VECTOR)
    </#if> 
    <#if CHIP_TYPE == "PIC18">
    if (userAddress < NEW_RESET_VECTOR)
    </#if>
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10U);
    }
</#if>

    // get that start of the page and the user data start address
    flashStartPageAddress = FLASH_PageAddressGet(userAddress);
    userDataStartOffset = FLASH_PageOffsetGet(userAddress);

<#if CHIP_TYPE == "PIC16">
    // read the whole page that contains the target address
    for (uint16_t offset = 0U; offset < PROGMEM_PAGE_SIZE; offset++)
    {
        writeBuffer[offset] = FLASH_Read(flashStartPageAddress + offset);
    }

    // loop over each byte in the data buffer one by one
    for (uint16_t byte = 0U; byte < frame.data_length; byte++)
    {
        // calculate the word offset based on the byte index
        uint16_t wordIndex = (byte / 2U);

        // calculate the parity of the current byte
        bool isEvenParity = ((byte % 2) == 0);

        /**
         * Note: The logic here works because we can only address whole words with PIC16.
         * If we want to adopt a byte access specific to the bootloader logic we will need
         * to rethink this function and processing loop design for PIC16F devices.
         * Even Parity -> data[byte] is a low byte
         * Odd Parity -> data[byte] is a high byte
         */
        if (isEvenParity == true)
        {
            writeBuffer[userDataStartOffset + wordIndex] = (writeBuffer[userDataStartOffset + wordIndex] & 0xFF00U) | frame.data[byte];
        }
        else
        {
            writeBuffer[userDataStartOffset + wordIndex] = (writeBuffer[userDataStartOffset + wordIndex] & 0x00FFU) | ((flash_address_t) frame.data[byte]) << 8U;
        }
    }
</#if>
<#if CHIP_TYPE == "PIC18" || DEVICE_TYPE == "AVR">
    // read the whole page that contains the address
    for (uint16_t offset = 0U; offset < BL_FRAME_DATA_SIZE; offset++)
    {
        writeBuffer[offset] = FLASH_Read(flashStartPageAddress + offset);
    }

    for (uint16_t userByte = 0U; userByte < frame.data_length; userByte++)
    {
        writeBuffer[userDataStartOffset + userByte] = frame.data[userByte];
    }
</#if>
    // ***** perform write action *****
<#if DEVICE_TYPE == "PIC">
    NVM_UnlockKeySet(unlockKey);
</#if>
    errorStatus = FLASH_PageErase(flashStartPageAddress);
<#if DEVICE_TYPE == "PIC">
    NVM_UnlockKeyClear();
</#if>
    if (errorStatus == NVM_OK)
    {
<#if DEVICE_TYPE == "PIC">
        NVM_UnlockKeySet(unlockKey);
</#if>
        errorStatus = FLASH_RowWrite(flashStartPageAddress, writeBuffer);
<#if DEVICE_TYPE == "PIC">
        NVM_UnlockKeyClear();
</#if>
    }

    frame.data[0] = (errorStatus == NVM_OK) ? COMMAND_SUCCESS : COMMAND_PROCESSING_ERROR;

<#if (CHIP_TYPE != "TINY")>
    NVM_StatusClear();
</#if>
    return (10U);
}


/************************************************************************************************
 * Erase Application Flash Space
 *        Cmd--- Length----- Keys------- Address------------------------- Data ------------------
<#if DEVICE_TYPE == "PIC">
 * In:   [|0x03 | DATALEN_L | DATALEN_L | 0x55 | 0xAA | ADDR_L | ADDR_H | ADDR_U | ADDR_E|]
<#else>
 * In:   [|0x03 | DATALEN_L | DATALEN_L | 0x00 | 0x00 | ADDR_L | ADDR_H | ADDR_U | ADDR_E|]
</#if>
 * OUT:  [|0x03 | DATALEN_L | DATALEN_L | KEY_L | KEY_H | ADDR_L | ADDR_H | ADDR_U | ADDR_E | CMD_STATUS|]
 ************************************************************************************************
 */
static uint8_t BL_EraseFlash(void)
{
    nvm_status_t errorStatus = NVM_OK;
    flash_address_t address;

<#if DEVICE_TYPE == "PIC">
    uint16_t unlockKey;
    unlockKey = (((uint16_t) frame.EE_key_2) << 8U)
            | (uint16_t) frame.EE_key_1;
<#if CHIP_TYPE == "PIC18">
    address = ((((flash_address_t) frame.address_U) << 16U)
            | (((flash_address_t) frame.address_H) << 8U)
            | (flash_address_t) frame.address_L);
</#if>
<#if CHIP_TYPE == "PIC16">
    address = (((flash_address_t)frame.address_H) << 8U)
            | (flash_address_t)frame.address_L;
</#if>

    // Fail if the given unlock key is incorrect
    if (unlockKey != UNLOCK_KEY)
    {
        frame.data[0] = COMMAND_PROCESSING_ERROR;
<#if (CHIP_TYPE != "TINY")>
        NVM_StatusClear();
</#if>
        return (10U);
    }
</#if>
<#if DEVICE_TYPE == "AVR">
<#if AVR_PAGE == "Expanded">
    address = ( 
            ((flash_address_t)frame.address_E << 24U) 
            | ((flash_address_t)frame.address_U << 16U) 
            | ((flash_address_t)frame.address_H << 8U) 
            | frame.address_L
            );
<#else>
    address = ( 
            ((flash_address_t)frame.address_H << 8U) 
                | frame.address_L
            );
</#if>
</#if>

    // Fail if the given address is not on a page boundry
    if (FLASH_PageOffsetGet(address) > (flash_address_t) 0U)
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
<#if (CHIP_TYPE != "TINY")>
        NVM_StatusClear();
</#if>
        return (10U);
    }

<#if DEVICE_TYPE == "PIC">
<#if (SELF_PROTECTION == "Enabled")>
    // Fail if the given address is inside the boot space
<#if CHIP_TYPE == "PIC16" >
    if ((address & (flash_address_t) (0x7FFFU & (~LAST_WORD_MASK))) < (flash_address_t) NEW_RESET_VECTOR)
</#if>
<#if CHIP_TYPE == "PIC18" >
    if (address < (flash_address_t) NEW_RESET_VECTOR)
</#if>
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        NVM_StatusClear();
        return (10U);
    }
</#if>

    for (uint16_t i = 0U; i < frame.data_length; i++)
    {
        NVM_UnlockKeySet(unlockKey);
        errorStatus = FLASH_PageErase(address);
        NVM_UnlockKeyClear();

        address += PROGMEM_PAGE_SIZE;

        if (errorStatus == NVM_ERROR)
        {
            break;
        }
    }
</#if>
<#if DEVICE_TYPE == "AVR">
<#if (ERASE_CONTROL == "FullErase")>
    if( address != START_OF_APP ) {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10U);
    }
    
    while (address < FLASHEND)
    {
        errorStatus = FLASH_PageErase(address);
        if (errorStatus != NVM_OK)
        {
            break;
        }
        address += PROGMEM_PAGE_SIZE;    
    }
</#if>
<#if (ERASE_CONTROL == "BlockBased")>
    if ((address  < START_OF_APP) || (address > FLASHEND) )
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (BL_HEADER + 1U);
    }
    
<#if (AVR_PAGE == "Expanded")>
    for (uint32_t i = 0U; i < frame.data_length; i ++)
</#if>
<#if (AVR_PAGE == "Classic")>
    for (uint8_t i = 0U; i < frame.data_length; i ++)
</#if>
    {
        errorStatus = FLASH_PageErase(address);
        if (errorStatus != NVM_OK)
        {
            break;
        }
        address += PROGMEM_PAGE_SIZE;
    }
</#if>
</#if>

    frame.data[0] = (errorStatus == NVM_OK) ? COMMAND_SUCCESS : COMMAND_PROCESSING_ERROR;
<#if (CHIP_TYPE != "TINY")>
    NVM_StatusClear();
</#if>
    return (10U);
}

<#if (SUPPORT_EE_DATA_READ == "Enabled")>
/** 
 * @todo Finish the API comments here
 * @example
 * ************************************************************************************************
 * Read data from the EEPROM area of memory
 *        Cmd--- Length----- Keys------- Address------------------------- Data ------------------
<#if DEVICE_TYPE == "PIC">
 * In:   [|0x04 | DATALEN_L | DATALEN_H | 0x55 | 0xAA | ADDR_L | ADDR_H | ADDR_U | ADDR_E|]
<#else>
 * In:   [|0x04 | DATALEN_L | DATALEN_H | 0x00 | 0x00 | ADDR_L | ADDR_H | ADDR_U | ADDR_E|]
</#if>
 * OUT:  [|0x04 | DATALEN_L | DATALEN_H | KEY_L | KEY_H | ADDR_L | ADDR_H | ADDR_U | ADDR_E | CMD_STATUS | Data |.. | data |]
 * ************************************************************************************************
 */
static uint16_t BL_ReadEEData(void)
{
    eeprom_address_t address;
<#if DEVICE_TYPE == "AVR">
    address = ( ( (eeprom_address_t) frame.address_H << 8U) | frame.address_L); 
</#if>  
<#if CHIP_TYPE == "PIC16">
    address = ((eeprom_address_t)frame.address_H )<< 8U | frame.address_L;
</#if> 
<#if CHIP_TYPE == "PIC18">
    address = (((eeprom_address_t) frame.address_U) << 16U)
            | (((eeprom_address_t) frame.address_H) << 8U)
            | (eeprom_address_t) frame.address_L;
</#if> 

    if(!((address >= EEPROM_START_ADDRESS_U) && (address < (EEPROM_START_ADDRESS_U + EEPROM_SIZE_U))))
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return 10U;
    }

    // Prevent any read operation that exceeds the data buffer size
    if( frame.data_length > BL_FRAME_DATA_SIZE ) {
        frame.data[0] = COMMAND_OVERLOAD_ERROR;
        return (10U);
    }

    for (uint16_t i = 0U; i < frame.data_length; i++)
    {
        frame.data[i + 1U] = EEPROM_Read(address);
        address++;
    }
    frame.data[0] = (NVM_StatusGet() == NVM_OK) ? COMMAND_SUCCESS : COMMAND_PROCESSING_ERROR;
    return (frame.data_length + 10U);
}
</#if>

<#if (SUPPORT_EE_DATA_WRITE == "Enabled")>
/** 
 * @todo Finish the API comments here
 * @example
 * ************************************************************************************************
 * Write data into the EEPROM area of memory
 *        Cmd--- Length----- Keys------- Address------------------------- Data ------------------
<#if DEVICE_TYPE == "PIC">
 * In:   [|0x05 | DATALEN_L | DATALEN_H | 0x55 | 0xAA | ADDR_L | ADDR_H | ADDR_U | ADDR_E | Data |.. | data |]
<#else>
 * In:   [|0x05 | DATALEN_L | DATALEN_H | 0x00 | 0x00 | ADDR_L | ADDR_H | ADDR_U | ADDR_E | Data |.. | data |]
</#if>
 * OUT:  [|0x05 | DATALEN_L | DATALEN_H | KEY_L | KEY_H | ADDR_L | ADDR_H | ADDR_U | ADDR_E | CMD_STATUS|]
 * ************************************************************************************************
 */
static uint8_t BL_WriteEEData(void)
{
    eeprom_address_t address;

    // Prevent any write operation that exceeds the data buffer size
    if( frame.data_length > BL_FRAME_DATA_SIZE ) {
        frame.data[0] = COMMAND_OVERLOAD_ERROR;
        return (10U);
    }

<#if CHIP_TYPE == "PIC16">
    uint16_t unlockKey = ((uint16_t)frame.EE_key_2) << 8U | frame.EE_key_1;
    
    address = ((eeprom_address_t)frame.address_H ) << 8U | frame.address_L;
    
    if(!((address >= EEPROM_START_ADDRESS_U) && (address < (EEPROM_START_ADDRESS_U + EEPROM_SIZE_U))))
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return 10U;
    }

    for (uint16_t i = 0U; i < frame.data_length; i++)
    {
        NVM_UnlockKeySet(unlockKey);
        EEPROM_Write(address++, (eeprom_data_t)frame.data[i]);
        while(NVM_IsBusy())
        {
        }
        NVM_UnlockKeyClear();
            
        if(NVM_StatusGet() != NVM_OK)
        {
            NVM_StatusClear();
            frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
            return (BL_HEADER + 1U);
        }
    }
    frame.data[0] = COMMAND_SUCCESS;
    return (BL_HEADER + 1U);
</#if> 
<#if DEVICE_TYPE == "AVR">
    address = ( ( (eeprom_address_t)frame.address_H << 8U) | frame.address_L);

    if(!((address >= EEPROM_START_ADDRESS_U) && (address < (EEPROM_START_ADDRESS_U + EEPROM_SIZE_U))))
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return 10U;
    }
	
    for(uint16_t i = 0; i < frame.data_length; i++)
    {
        EEPROM_Write(address++, frame.data[i]);
        while(EEPROM_IsBusy())
        {
        }
        if(NVM_StatusGet() != NVM_OK) 
        {
            frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
            return (BL_HEADER + 1U);
        }
    }
    frame.data[0] = COMMAND_SUCCESS;
    return (BL_HEADER + 1U);
</#if> 
<#if CHIP_TYPE == "PIC18">

    uint16_t unlockKey = (((uint16_t) frame.EE_key_2) << 8U) | (uint16_t) frame.EE_key_1;
    
    address = (((eeprom_address_t) frame.address_U) << 16U)
            | (((eeprom_address_t) frame.address_H) << 8U)
            | (eeprom_address_t) frame.address_L;

    if(!((address >= EEPROM_START_ADDRESS_U) && (address < (EEPROM_START_ADDRESS_U + EEPROM_SIZE_U))))
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return 10U;
    }
    
    for (uint16_t i = 0U; i < frame.data_length; i++)
    {
        NVM_UnlockKeySet(unlockKey);
        EEPROM_Write(address++, frame.data[i]);
        while (NVM_IsBusy())
        {
        }
        NVM_UnlockKeyClear();

        if (NVM_StatusGet() != NVM_OK)
        {
            NVM_StatusClear();
            frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
            return (BL_HEADER + 1U);
        }
    }
    frame.data[0] = COMMAND_SUCCESS;
    return (BL_HEADER + 1U);
</#if> 
}
</#if>

<#if SUPPORT_CONFIG_READ == "Enabled">
<#if CHIP_TYPE == "PIC16">
// **************************************************************************************
// Read Config Words
// In:    [|0x06 | DataLengthL | unused | unused | unused | ADDRL | ADDRH | ADDRU | unused |...]
// OUT:   [9 byte header + 4 bytes config1 + config 2]
// *****************************************************************************
</#if>
<#if CHIP_TYPE == "PIC18">
// *****************************************************************************
// Read Config Words
//        Cmd     Length-----              Address---------------  Data ---------
// In:   [|0x06 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00|]
// OUT:  [|0x06 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | Data |.. | data |]
// *****************************************************************************
</#if>
<#if DEVICE_TYPE == "AVR">
// **************************************************************************************
// Request Read device FUSE setting using bootloader
// In:    [|0x06 | DataLengthL | unused | unused | unused | ADDRL | ADDRH | ADDRU | unused |...]
// OUT:    [9 byte header + data ]
// **************************************************************************************
</#if>
static uint16_t BL_ReadConfig(void)
{
<#if DEVICE_TYPE == "PIC">
    configuration_address_t configurationAddress;
</#if>
<#if CHIP_TYPE == "PIC16">
    configuration_data_t readWord;
</#if>
    uint8_t dataIndex = 1;
<#if CHIP_TYPE == "PIC16">
    configurationAddress = (((uint16_t)frame.address_H) << 8U) | frame.address_L;
        
    for (uint8_t  i = 0U; i < frame.data_length; i += 2U)
    {
        readWord = CONFIGURATION_Read(configurationAddress);
        configurationAddress++;
        frame.data[dataIndex++] = (uint8_t)readWord;
        frame.data[dataIndex++] = (uint8_t)(readWord >> 8U);
    }
    frame.data[0] = (NVM_StatusGet() == NVM_OK)? COMMAND_SUCCESS: COMMAND_PROCESSING_ERROR;
    
    NVM_StatusClear();
    return (BL_HEADER + dataIndex);
</#if>

<#if CHIP_TYPE == "PIC18">
    configurationAddress = (((uint32_t) frame.address_U) << 16U)
            | (((uint32_t) frame.address_H) << 8U)
            | (uint32_t) frame.address_L;

<#if (SELF_PROTECTION == "Enabled")>
    if (configurationAddress < (configuration_address_t) NEW_RESET_VECTOR)
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10U);
    }
</#if>
    for (uint8_t i = 0; i < frame.data_length; i++)
    {
        frame.data[dataIndex] = EEPROM_Read(configurationAddress);
        configurationAddress++;
        dataIndex++;
    }

    frame.data[0] = (NVM_StatusGet() == NVM_OK)? COMMAND_SUCCESS: COMMAND_PROCESSING_ERROR;
    NVM_StatusClear();

    return (BL_HEADER + dataIndex);
</#if>

<#if DEVICE_TYPE == "AVR">
    frame.data[dataIndex++] = FUSE.WDTCFG;
    frame.data[dataIndex++] = FUSE.BODCFG;
    frame.data[dataIndex++] = FUSE.OSCCFG;
    
    frame.data[dataIndex++] = FUSE.reserved_1[0];
<#if (FUSE == "Reserved_One_Byte")>
    frame.data[dataIndex++] = FUSE.TCD0CFG;
</#if>
<#if (FUSE == "Reserved_Two_Byte")>
    frame.data[dataIndex++] = FUSE.reserved_1[1];
</#if>

    frame.data[dataIndex++] = FUSE.SYSCFG0;
    frame.data[dataIndex++] = FUSE.SYSCFG1;
    
<#if (BOOT_BLOCK_SIZE_U == "512bytes")>
    frame.data[dataIndex++] = FUSE.CODESIZE;
    frame.data[dataIndex++] = FUSE.BOOTSIZE;
</#if>
<#if (BOOT_BLOCK_SIZE_U == "256bytes")>
    frame.data[dataIndex++] = FUSE.APPEND;
    frame.data[dataIndex++] = FUSE.BOOTEND;
</#if> 

    frame.data[0] = (NVM_StatusGet() == NVM_OK)? COMMAND_SUCCESS: COMMAND_PROCESSING_ERROR;
<#if (CHIP_TYPE != "TINY")>
    NVM_StatusClear();
</#if>
    return (BL_HEADER + dataIndex);     // 9 byte header + data 
</#if>
}
</#if>

<#if SUPPORT_CONFIG_WRITE == "Enabled">
// **************************************************************************************
// Write Config Words
//        Cmd     Length-----              Address---------------  Data ---------
// In:   [|0x07 | 0x00 | 0x00 | 0x55 | 0xAA | 0x00 | 0x00 | 0x00 | 0x00 | Data |.. | data |]
// OUT:  [|0x07 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00|]
// **************************************************************************************
static uint8_t BL_WriteConfig(void)
{
    configuration_address_t configurationAddress;
    uint16_t unlockKey;
<#if CHIP_TYPE == "PIC16">
    configuration_data_t dataWord;
</#if>
<#if CHIP_TYPE == "PIC18">
    uint8_t configurationByte = 0U;
</#if>

<#if CHIP_TYPE == "PIC16">
    configurationAddress = (((configuration_address_t)frame.address_H) << 8U) | frame.address_L;
</#if>
<#if CHIP_TYPE == "PIC18">
    configurationAddress = (((configuration_address_t) frame.address_U) << 16U)
            | (((configuration_address_t) frame.address_H) << 8U)
            | (configuration_address_t) frame.address_L;
</#if>

    unlockKey = (((uint16_t) frame.EE_key_2) << 8U)
            | (uint16_t) frame.EE_key_1;

<#if (SELF_PROTECTION == "Enabled")>
<#if CHIP_TYPE == "PIC16">
    if (configurationAddress < (configuration_address_t) NEW_RESET_VECTOR)
</#if>
<#if CHIP_TYPE == "PIC18">
    if (configurationAddress < (configuration_address_t) NEW_RESET_VECTOR)
</#if>
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10U);
    }
</#if>

<#if CHIP_TYPE == "PIC16">
    if( unlockKey != UNLOCK_KEY ) {
        frame.data[0] = ERROR_INVALID_COMMAND;
        return (10U);
    }
        
    for (uint8_t  i = 0U; i < frame.data_length; i += 2U)
    {
        dataWord = (((configuration_data_t) frame.data[i + 1U]) << 8U) | frame.data[i];
        
        // Do the write sequence
        NVM_UnlockKeySet(unlockKey);
        CONFIGURATION_Write(configurationAddress, dataWord);
        NVM_UnlockKeyClear();
        while (NVM_IsBusy())
        {
        }
    }
    frame.data[0] = (NVM_StatusGet() == NVM_OK)? COMMAND_SUCCESS: COMMAND_PROCESSING_ERROR;
    
    NVM_StatusClear();
</#if>
<#if CHIP_TYPE == "PIC18">
    NVM_UnlockKeySet(unlockKey);

    for (uint8_t i = 0U; i < frame.data_length; i++)
    {

        configurationByte = EEPROM_Read(configurationAddress);
        configurationByte = frame.data[i];
        EEPROM_Write(configurationAddress, configurationByte);

        ++configurationAddress;
    }

    NVM_UnlockKeyClear();

    frame.data[0] = (NVM_StatusGet() == NVM_OK)? COMMAND_SUCCESS: COMMAND_PROCESSING_ERROR;
    
    NVM_StatusClear();
</#if>
    return (10U);
}
</#if>

<#if SUPPORT_CHECKSUM == "Enabled">
// **************************************************************************************
// Calculate Checksum
// In:	[|0x08 | DataLengthL | DataLengthH | unused | unused | ADDRL | ADDRH | ADDRU | unused |...]
<#if DEVICE_TYPE == "PIC">
// OUT:	[9 byte header + ChecksumL + ChecksumH]
<#else>
// OUT:    [9 byte header + CRC_STATUS]
</#if>
// **************************************************************************************

static uint8_t BL_CalcChecksum(void)
{
<#if DEVICE_TYPE == "PIC">
    flash_address_t address;
#if PROGMEM_SIZE > 0x10000
    uint32_t i;
    uint32_t length = frame.data_length;
    length += ((uint32_t) frame.EE_key_1) << 16U;
#else
    uint16_t i;
    uint16_t length = frame.data_length;
#endif
<#if CHIP_TYPE == "PIC16">
    flash_data_t    wordRead;

    address = (((flash_address_t)frame.address_H) << 8U)
            |   (flash_address_t)frame.address_L;
<#else>
    address = (((flash_address_t) frame.address_U) << 16U)
            | (((flash_address_t) frame.address_H) << 8U)
            | (flash_address_t) frame.address_L;
</#if>

    if(address < START_OF_APP)
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10U);
    }
    
    uint16_t checkSum = 0U;

    for (i = 0U; i < length; i += 2U)
    {
<#if CHIP_TYPE == "PIC18">
        checkSum += (uint16_t) FLASH_Read(address++);
        checkSum += ((uint16_t) FLASH_Read(address++)) << 8U;
<#else>
        wordRead = FLASH_Read(address);
        checkSum += wordRead & 0x00FFU;
        checkSum += wordRead & 0xFF00U;
        ++address;
</#if>
    }

    frame.data[0] = (uint8_t) (checkSum & 0x00FFU);
    frame.data[1] = (uint8_t) ((checkSum & 0xFF00U) >> 8U);
    return (11U);
</#if>
<#if DEVICE_TYPE == "AVR">
    // not implemented yet
</#if>
}
</#if>

<#if VERIFY == "STATUS_BYTE_VERIFICATION">
/**
 * @ingroup 8bit_bootloader
 * @brief   Sets the status byte value of the bootloader in FLASH.
 * @param   None.
 * @return  The total length of the packet being passed back to the host.
 * @example
 * ************************************************************************************************
 * Write Application Status to Predefined Flash Address
 *        Cmd--- Length----- Keys------- Address------------------------- Data ------------------
<#if DEVICE_TYPE == "PIC">
 * In:   [|0x0B | 0x01 | 0x00 | 0x55 | 0xAA | ADDR_L | ADDR_H | ADDR_U | ADDR_E | ${APP_VALID_VALUE_U}|]
<#else>
 * In:   [|0x0B | 0x01 | 0x00 | 0x00 | 0x00 | ADDR_L | ADDR_H | ADDR_U | ADDR_E | ${APP_VALID_VALUE_U}|]
</#if>
 * OUT:  [|0x0B | 0x01 | 0x00 | 0x00 | 0x00 | ADDR_L | ADDR_H | ADDR_U | ADDR_E | CMD_STATUS|]
 * ************************************************************************************************
 */
static uint8_t BL_WriteAppStatus(void)
{
    nvm_status_t errorStatus = NVM_OK;
    flash_data_t userStatusByte = (flash_data_t) frame.data[0];
    flash_data_t writeBuff[PROGMEM_PAGE_SIZE];
<#if DEVICE_TYPE == "PIC">
    uint16_t unlockKey = (((uint16_t) frame.EE_key_2) << 8U)
            | (uint16_t) frame.EE_key_1;
</#if>

<#if DEVICE_TYPE == "AVR">
<#if AVR_PAGE == "Expanded">
    flash_address_t userAddress = ( 
            ((flash_address_t)frame.address_E << 24U) 
            | ((flash_address_t)frame.address_U << 16U) 
            | ((flash_address_t)frame.address_H << 8U) 
            | frame.address_L
            );
<#else>
    flash_address_t userAddress = ( 
            (((flash_address_t)frame.address_H << 8U) 
                | frame.address_L
            ));
</#if>
</#if>

<#if CHIP_TYPE == "PIC18">
    flash_address_t userAddress = (((flash_address_t) frame.address_U) << 16U)
            | (((flash_address_t) frame.address_H) << 8U)
            | (flash_address_t) frame.address_L;
</#if>
<#if CHIP_TYPE == "PIC16">
    flash_address_t userAddress = (((flash_address_t)frame.address_H) << 8U)
            |   (flash_address_t)frame.address_L;
</#if>

    // Validate the target address
    if ((STATUS_ADDRESS ^ userAddress) != 0U)
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (BL_HEADER + 1U);
    }

<#if DEVICE_TYPE == "PIC">
    if (unlockKey != UNLOCK_KEY)
    {
        frame.data[0] = COMMAND_PROCESSING_ERROR;
        return (BL_HEADER + 1U);
    }
</#if>
    flash_address_t startOfPageRow = FLASH_PageAddressGet((flash_address_t) STATUS_ADDRESS);
    flash_address_t statusOffset = FLASH_PageOffsetGet((flash_address_t) STATUS_ADDRESS);

    // loop over row to preserve the page data
    for (uint16_t i = 0U; i < PROGMEM_PAGE_SIZE; i++)
    {
        writeBuff[i] = FLASH_Read(startOfPageRow + i);
    }

    writeBuff[statusOffset] = userStatusByte;
    
<#if DEVICE_TYPE == "PIC">
    NVM_UnlockKeySet(unlockKey);
</#if>
    errorStatus = FLASH_PageErase(startOfPageRow);
<#if DEVICE_TYPE == "PIC">
    NVM_UnlockKeyClear();
</#if>
    if (errorStatus == NVM_OK)
    {
<#if DEVICE_TYPE == "PIC">
        NVM_UnlockKeySet(unlockKey);
</#if>
        errorStatus = FLASH_RowWrite(startOfPageRow, writeBuff);
<#if DEVICE_TYPE == "PIC">
        NVM_UnlockKeyClear();
</#if>
    }

    frame.data[0] = (errorStatus == NVM_OK) ? COMMAND_SUCCESS : COMMAND_PROCESSING_ERROR;
    return (BL_HEADER + 1U); // 9 byte header + Status
}
</#if>

<#if DEVICE_TYPE == "AVR">
// **************************************************************************************
// Reset
// In:   [|0x09 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00|]
// OUT:    [9 byte header] + [ChecksumL + ChecksumH]
// **************************************************************************************
static uint8_t BL_Reset(void)
{
    uint8_t dataIndex = 1U; 
    
    resetPending = true;
    return (BL_HEADER + dataIndex);
}
// *****************************************************************************
</#if>