/**
 *
 * @file bl_pic16f1_bootload.c
 *
 * @ingroup generic_bootloader_8bit
 *
 * @brief This source file provides the implementation of the APIs for the 8-bit Bootloader library
 *
 * @version BOOTLOADER Driver Version 3.0.0
 */

${disclaimer}

//
// Memory Map
//   -----------------
//   |    0x0000     |   Reset vector
//   |               |
//   |    0x0004     |   Interrupt vector
//   |               |
//   |               |
//   |  Boot Block   |   (this program)
//   |               |
//   |    ${NEW_RESET_VECTOR_U}     |   Re-mapped Reset Vector
//   |    ${NEW_INTERRUPT_VECTOR_LOW_U}     |   Re-mapped High Priority Interrupt Vector
//   |               |
//   |       |       |
//   |               |
//   |  Code Space   |   User program space
//   |               |
//   |       |       |
//   |               |
//   |    0x3FFF     |
//   -----------------


// *****************************************************************************
#include <stdbool.h>
#include "../bl_bootload.h"
#include "../bl_communication_interface.h"
// *****************************************************************************
static uint8_t  BL_GetVersionData(void);
static uint8_t  BL_ReadFlash(void);
static uint8_t  BL_WriteFlash(void);
static uint8_t  BL_EraseFlash(void);
static uint8_t  BL_ReadEEData(void);
static uint8_t  BL_WriteEEData(void);
static uint8_t  BL_ReadConfig(void);
static uint8_t  BL_WriteConfig(void);
static uint8_t  BL_CalcChecksum(void);
static void     BL_StartWrite(void);
static void     BL_RunBootloader(void);
static uint8_t  BL_ProcessBootBuffer (void);
static bool     BL_BootloadRequired (void);
static void 	BL_CheckDeviceReset (void);
// *****************************************************************************
static bool resetPending = false;

// Force variables into Unbanked for 1-cycle accessibility 
static uint8_t EE_Key_1    __at(0x70);
static uint8_t EE_Key_2    __at(0x71);

frame_t  frame;

// *****************************************************************************
// The bootloader code does not use any interrupts.
// However, the application code may use interrupts.
// The interrupt vector on a PIC16F is located at
// address 0x0004. 
// The following function will be located
// at the interrupt vector and will contain a jump to
// the new application interrupt vector

    asm ("psect  intentry,global,class=CODE,delta=2");
    asm ("pagesel " str(NEW_INTERRUPT_VECTOR));
    asm ("GOTO " str(NEW_INTERRUPT_VECTOR));

void BL_Initialize (void)
{
    EE_Key_1 = 0;
    EE_Key_2 = 0;
<#if (BOOTLOADER_RUNNING_IO == "Enabled")>
	BL_INDICATOR_OFF();
</#if>
    if (BL_BootloadRequired() == true)
    {
        BL_INDICATOR_ON();
        BL_RunBootloader ();     // generic comms layer
    }
    STKPTR = 0x1F;
    BSR = 0;
<#if (BOOTLOADER_RUNNING_IO == "Enabled")>
    BL_INDICATOR_OFF();
</#if>
    asm ("pagesel " str(NEW_RESET_VECTOR));
    asm ("goto  "  str(NEW_RESET_VECTOR));
}

static void BL_RunBootloader(void)
{
    volatile uint16_t  index;
    uint16_t  messageLength;
    uint8_t ch;

    while (1)
    {
        while(BL_CommunicationModuleIsReady() != true); // wait for last byte to shift out before
                                                        // starting autobaud.
        BL_CheckDeviceReset ();  // Response has been sent.  Check to see if a reset was requested

        BL_CommunicationModuleInit();

        // Read and parse the data
        index = 0;       // Point to the buffer
        messageLength = 9;  // message has 9 bytes of overhead (Opcode + Length + Address)

        while (index < messageLength)
        {
            BL_CommunicationModuleRead(&ch,1);
            frame.buffer [index ++] = ch;
            if (index == 4)
            {
                if ((frame.command == WRITE_FLASH)
                 || (frame.command == WRITE_EE_DATA)
                 || (frame.command == WRITE_CONFIG))
                {
                    messageLength += frame.data_length;
                }
            }
        }

        messageLength = BL_ProcessBootBuffer ();

        // Send the data buffer back
        BL_CommunicationModuleWrite(frame.buffer, messageLength);       
    }
}

// *****************************************************************************
static bool BL_BootloadRequired(void)
{
<#if IO_PIN_ENTRY == "Enabled">
// ******************************************************************
//  Check an IO pin to force entry into bootloader
// ******************************************************************

//#info  "You may need to add additional delay here between enabling weak pullups and testing the pin."
    for (uint8_t i = 0; i != 0xFF; i++) NOP();

    if (IO_PIN_ENTRY_GetInputValue() == IO_PIN_ENTRY_RUN_BL)
    {
        return (true);
    }
</#if>

    if (BL_bootVerify() == false)
    {
        return (true);
    }

    return (false);
}


// *****************************************************************************
static uint8_t  BL_ProcessBootBuffer(void)
{
    uint8_t len;

    EE_Key_1 = frame.EE_key_1;
    EE_Key_2 = frame.EE_key_2;

// ***********************************************
// Test the command field and sub-command.
    switch (frame.command)
    {
    case    READ_VERSION:
        len = BL_GetVersionData();
        break;
<#if (SUPPORT_FLASH_READ == "Enabled")>
    case    READ_FLASH:
        len = BL_ReadFlash();
        break;
</#if>
    case    WRITE_FLASH:
        len = BL_WriteFlash();
        break;
    case    ERASE_FLASH:
        len = BL_EraseFlash();
        break;
<#if (SUPPORT_EE_DATA_READ == "Enabled")>
    case    READ_EE_DATA:
        len = BL_ReadEEData();
        break;
</#if>
<#if (SUPPORT_EE_DATA_WRITE == "Enabled")>
    case    WRITE_EE_DATA:
        len = BL_WriteEEData();
        break;
</#if>
    case    READ_CONFIG:
        len = BL_ReadConfig();
        break;
<#if SUPPORT_CONFIG_WRITE == "Enabled">
    case    WRITE_CONFIG:
        len = BL_WriteConfig();
        break;
</#if>
    case    CALC_CHECKSUM:
        len = BL_CalcChecksum();
        break;
    case    RESET_DEVICE:
        frame.data[0] = COMMAND_SUCCESS;
        resetPending = true;
        len = 10;
        break;
    default:
        frame.data[0] = ERROR_INVALID_COMMAND;
        len = 10;
    }
    return (len);
}

// **************************************************************************************
// Commands
//
//        Cmd     Length----------------   Address---------------
// In:   [|0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 |]
// OUT:  [|0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | 0x00 | VERL | VERH |]
static uint8_t  BL_GetVersionData(void)
{
    frame.data[0] = MINOR_VERSION;
    frame.data[1] = MAJOR_VERSION;
    frame.data[2] = 0;       // Max packet size (256)
    frame.data[3] = 1;
    frame.data[4] = 0;
    frame.data[5] = 0;
    ${EEADRL_SFR} = 0x06;               // Get device ID
    ${EEADRH_SFR} = 0x00;
    ${EECON1_SFR}bits.${CFGS_BIT} = 1;
    ${EECON1_SFR}bits.RD = 1;
    NOP();
    NOP();
    frame.data[6] = ${EEDATL_SFR};
    frame.data[7] = ${EEDATH_SFR};
    frame.data[8] = 0;
    frame.data[9] = 0;
    frame.data[10] = PROGMEM_PAGE_SIZE_LOW_BYTE;
    frame.data[11] = PROGMEM_PAGE_SIZE_HIGH_BYTE;

    ${EEADRL_SFR} = 0x00;
    ${EEADRH_SFR} = 0x00;
    ${EECON1_SFR}bits.${CFGS_BIT} = 1;
    for (uint8_t  i= 12; i < 16; i++)
    {
        ${EECON1_SFR}bits.RD = 1;
        NOP();
        NOP();
        frame.data[i] = ${EEDATL_SFR};
        if ((++ ${EEADRL_SFR}) == 0x00)
        {
            ++ ${EEADRH_SFR};
        }
    }

    return  25;   // total length to send back 9 byte header + 16 byte payload
}

<#if (SUPPORT_FLASH_READ == "Enabled")>
// **************************************************************************************
// Read Flash
// In:	[|0x01 | DLEN | ADDRL | ADDRH | ADDRU |]
// OUT:	[|0x01 | DLEN | ADDRL | ADDRH | ADDRU | DATA |...]
static uint8_t BL_ReadFlash(void)
{
    flash_address_t address;
    flash_data_t    wordRead;

    address = (((flash_address_t)frame.address_H) << 8) 
            | (flash_address_t)frame.address_L;

    for (uint16_t i = 0; i < frame.data_length; i += 2)
    {
        wordRead = FLASH_Read(address);
        frame.data[i] = wordRead & 0x00FF;
        frame.data[i + 1] = wordRead & 0xFF00; 
        ++address;
    }

    return (frame.data_length + 9);
}
</#if>

// **************************************************************************************
// Write Flash
// In:	[|0x02 | DLENBLOCK | 0x55 | 0xAA | ADDRL | ADDRH | ADDRU | DATA |...]
// OUT:	[|0x02 |]
static uint8_t BL_WriteFlash(void)
{
    flash_address_t address;
    nvm_status_t errorStatus = NVM_OK;
    flash_address_t flashStartPageAddress;
    flash_data_t    wordBuffer[PROGMEM_PAGE_SIZE];
    uint8_t         dataIndex = 0; 
    uint16_t unlockKey;

    address = (((flash_address_t)frame.address_H) << 8)
            | (flash_address_t)frame.address_L;

    unlockKey = (((uint16_t)EE_Key_2) << 8)
            | (uint16_t)EE_Key_1;

    NVM_StatusClear();
<#if (SELF_PROTECTION == "Enabled")>

    if ((address & 0x7FFF) < NEW_RESET_VECTOR)
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10);
    }
</#if>

    if (frame.data_length < BL_FRAME_DATA_SIZE)
    {
        for(uint8_t i = frame.data_length; i < BL_FRAME_DATA_SIZE; i++)
        {
            frame.data[i] = 0xFF;
        }
    }
    
    for (uint8_t i = 0; i < PROGMEM_PAGE_SIZE; i++)
    {
        dataIndex = i * 2;
        wordBuffer[i] = (((flash_data_t)frame.data[dataIndex + 1]) << 8)
                        | (flash_data_t)frame.data[dataIndex];
    }

    NVM_UnlockKeySet(unlockKey);

    flashStartPageAddress = FLASH_PageAddressGet(address);
    errorStatus = FLASH_RowWrite(flashStartPageAddress, wordBuffer);

    NVM_UnlockKeyClear();

    if (errorStatus == NVM_ERROR)
    {
        NVM_StatusClear();
    }

    frame.data[0] = errorStatus == NVM_OK ? COMMAND_SUCCESS : ERROR_INVALID_COMMAND;
    EE_Key_1 = 0x00;  // erase EE Keys
    EE_Key_2 = 0x00;
    return (10);
}

// **************************************************************************************
// Erase Program Memory
// Erases data_length rows from program memory
static uint8_t BL_EraseFlash(void)
{
    flash_address_t address;
    uint16_t unlockKey;
    nvm_status_t errorStatus = NVM_OK;

    address = (((flash_address_t)frame.address_H) << 8)
            | (flash_address_t)frame.address_L;

    unlockKey = (((uint16_t)EE_Key_2) << 8)
            | (uint16_t)EE_Key_1;
<#if (SELF_PROTECTION == "Enabled")>

    if ((address & (0x7FFF & (~LAST_WORD_MASK))) < NEW_RESET_VECTOR)
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10);
    }
</#if>

    NVM_UnlockKeySet(unlockKey);

    for (uint16_t i=0; i < frame.data_length; i++)
    {
        if ((((address >> 8)) & 0x7F) >= ((PROGMEM_SIZE & 0xFF00) >> 8))
        {
            frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
            return (10);
        }

        errorStatus = FLASH_PageErase(address);
        address += PROGMEM_PAGE_SIZE;

        if (errorStatus == NVM_ERROR)
        {
            break;
        }
    }

    NVM_UnlockKeyClear();
    
    frame.data[0]  = errorStatus == NVM_OK ? COMMAND_SUCCESS : ERROR_INVALID_COMMAND;
    frame.EE_key_1 = 0x00;  // erase EE Keys
    frame.EE_key_2 = 0x00;
    return (10);
}

<#if (SUPPORT_EE_DATA_READ == "Enabled")>
// **************************************************************************************
// Read EE Data
//
// In:	[|0x04 | DataLengthL | DataLengthH | unused | unused | <ADDRL | ADDRH | ADDRU | unused |...]
// OUT:	[|0x04 | DataLengthL | DataLengthH | unused | unused | ADDRL | ADDRH | ADDRU | unused | DATA |...]
//
static uint8_t BL_ReadEEData(void)
{
    ${EEADRL_SFR} = frame.address_L;
    ${EEADRH_SFR} = frame.address_H | 0x70;  // 0x70 needed for new style ee access
    ${EECON1_SFR} = 0x40;   // sets PGD bit for old style eeaccess
    for (uint8_t  i = 0; i < frame.data_length; i++)
    {
        ${EECON1_SFR}bits.RD = 1;
        NOP();
        NOP();
        frame.data[i] = ${EEDATL_SFR};
        ++ ${EEADRL_SFR};
    }
    return (frame.data_length+9);
}
</#if>
<#if (SUPPORT_EE_DATA_WRITE == "Enabled")>

// **************************************************************************************
// Write EE Data
// In:	[|0x05 | DLEN | ADDRL | ADDRH | 0x00 | DATA |...]
// OUT:	[|0x05 |]
static uint8_t  BL_WriteEEData(void)
{
    ${EEADRL_SFR} = frame.address_L;
    ${EEADRH_SFR} = frame.address_H | 0x70;
    ${EECON1_SFR} = 0x44;
    for (uint8_t i = 0; i < frame.data_length; i++)
    {
        ${EEDATL_SFR} = frame.data[i];
        BL_StartWrite ();
        while (${EECON1_SFR}bits.WR == 1);  // wait until previous write complete
        ++ ${EEADRL_SFR};
   }
    frame.data[0] = 0x01;
    return 10;
}
</#if>

// **************************************************************************************
// Read Config Words
// In:	[|0x06 | DataLengthL | unused | unused | unused | ADDRL | ADDRH | ADDRU | unused |...]
// OUT:	[9 byte header + 4 bytes config1 + config 2]
static uint8_t BL_ReadConfig(void)
{
    ${EEADRL_SFR} = frame.address_L;
    ${EEADRH_SFR} = frame.address_H;
    ${EECON1_SFR} = 0x40;      // can these be combined?
    for (uint8_t  i= 0; i < frame.data_length; i += 2)
    {

        ${EECON1_SFR}bits.RD = 1;
        NOP();
        NOP();
        frame.data[i]   = ${EEDATL_SFR};
        frame.data[i+1] = ${EEDATH_SFR};
        ++ ${EEADRL_SFR};
    }
    return (13);           // 9 byte header + 4 bytes config words
}
<#if SUPPORT_CONFIG_WRITE == "Enabled">
// **************************************************************************************
// Write Config Words
static uint8_t BL_WriteConfig(void)
{
    ${EEADRL_SFR} = frame.address_L;
    ${EEADRH_SFR} = frame.address_H;
<#if (SELF_PROTECTION == "Enabled")>
    {
        uint16_t  Address_Ptr;

        Address_Ptr = ((uint16_t) ${EEADRH_SFR}) << 8;
        Address_Ptr += ${EEADRL_SFR};
        if ((Address_Ptr & 0x7FFF) < NEW_RESET_VECTOR)
        {
            frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
            return (10);
        }
    }
</#if>
    ${EECON1_SFR} = 0xC4;       // Setup writes
    for (uint8_t  i = 0; i < frame.data_length; i += 2)
    {
        ${EEDATL_SFR} = frame.data[i];
        ${EEDATH_SFR} = frame.data[i+1];

        BL_StartWrite();
        if ((++ ${EEADRL_SFR}) == 0x00)
        {
            ++ ${EEADRH_SFR};
        }
    }
    frame.data[0] = COMMAND_SUCCESS;
    EE_Key_1 = 0x00;  // erase EE Keys
    EE_Key_2 = 0x00;
    return (10);
}
</#if>


<#if SUPPORT_CHECKSUM == "Enabled">
// **************************************************************************************
// Calculate Checksum
// In:	[|0x08 | DataLengthL | DataLengthH | unused | unused | ADDRL | ADDRH | ADDRU | unused |...]
// OUT:	[9 byte header + ChecksumL + ChecksumH]
static uint8_t BL_CalcChecksum(void)
{
    flash_address_t address;
    flash_data_t    wordRead;

    address = (((flash_address_t)frame.address_H) << 8)
            |   (flash_address_t)frame.address_L;

    uint16_t checkSum = 0;
    for (uint16_t i = 0; i < frame.data_length; i += 2)
    {
        wordRead = FLASH_Read(address);
        checkSum += wordRead & 0x00FF;
        checkSum += wordRead & 0xFF00;
        ++address;
     }
     frame.data[0] = (uint8_t) (checkSum & 0x00FF);
     frame.data[1] = (uint8_t)((checkSum & 0xFF00) >> 8);
     return (11);
}
</#if>

// *****************************************************************************
// Unlock and start the write or erase sequence.

static void BL_StartWrite(void)
{
    CLRWDT();
//    ${EECON2_SFR} = EE_Key_1;
//    ${EECON2_SFR} = EE_Key_2;
//    ${EECON1_SFR}bits.WR = 1;       // Start the write
// had to switch to assembly - compiler doesn't comprehend no need for bank switch
    asm ("movf " str(_EE_Key_1) ",w");
    asm ("movwf " str(BANKMASK(${EECON2_SFR})));
    asm ("movf  " str(_EE_Key_2) ",w");
    asm ("movwf " str(BANKMASK(${EECON2_SFR})));
    asm ("bsf  "  str(BANKMASK(${EECON1_SFR})) ",1");       // Start the write

    NOP();
    NOP();
    return;
}


// *****************************************************************************
// Check to see if a device reset had been requested.  We can't just reset when
// the reset command is issued.  Instead we have to wait until the acknowledgement
// is finished sending back to the host.  Then we reset the device.
static void BL_CheckDeviceReset(void)
{
    if (resetPending == true)
    {
<#if VERIFY == "STATUS_BYTE_VERIFICATION">
        // *************************************
        // Write the status of the bootloader at the device reset
        // which should occur after every reprogramming action
        // *************************************
        if( FLASH_Read(STATUS_ADDRESS) == 0x3FFF)
        {   
            flash_address_t flashStartPageAddress;
            nvm_status_t status = NVM_OK;
            flash_data_t wordBuff[PROGMEM_PAGE_SIZE];
            uint16_t unlockKey = (((uint16_t)EE_Key_2) << 8)
                                 | (uint16_t)EE_Key_1;
            
            flashStartPageAddress = FLASH_ErasePageAddressGet(STATUS_ADDRESS);
            
            for (uint8_t i = 0; i < PROGMEM_PAGE_SIZE; i++)
            {
                wordBuff[i] = FLASH_Read(flashStartPageAddress + i);
            }
            wordBuff[PROGMEM_PAGE_SIZE-1] = (uint16_t)APPLICATION_VALID;
            
            NVM_UnlockKeySet(unlockKey);
            status = FLASH_RowWrite(flashStartPageAddress, wordBuff);
            NVM_UnlockKeyClear();

            if (status == NVM_ERROR)
            {
                NVM_StatusClear();
            }
        }
</#if>

<#if (BOOTLOADER_RUNNING_IO == "Enabled")>
        BL_INDICATOR_OFF();
</#if>
        RESET();
    }
}

