/**
 *
 * @file bl_pic18f_bootload.c
 *
 * @ingroup generic_bootloader_8bit
 *
 * @brief This source file provides the implementation of the APIs for the 8-bit Bootloader library
 *
 * @version BOOTLOADER Driver Version 3.0.0
 */

${disclaimer}

// Memory Map
//   -----------------
//   |    0x0000     |   Reset vector
//   |               |
//   |    0x0008     |   High Priority Interrupt vector
//   |               |
//   |    0x0018     |   Low Priority Interrupt vector
//   |               |
//   |  Boot Block   |   (this program)
//   |               |
//   |    ${NEW_RESET_VECTOR_U}     |   Re-mapped Reset Vector (Actual address
//   |    ${NEW_INTERRUPT_VECTOR_HIGH_U}     |   Re-mapped High Priority Interrupt Vector
//   |    ${NEW_INTERRUPT_VECTOR_LOW_U}     |   Re-mapped Low Priority Interrupt Vector
//   |       |       |
//   |               |
//   |  Code Space   |   User program space
//   |               |
//   |       |       |
//   |               |
//   | End of Flash  |
//   -----------------
//
//
//
//
// *****************************************************************************


// *****************************************************************************
#include <stdbool.h>
#include "../bl_bootload.h"
#include "../bl_communication_interface.h"

// *****************************************************************************
static uint16_t BL_GetVersionData(void);
<#if (SUPPORT_FLASH_READ == "Enabled")>
static uint16_t BL_ReadFlash(void);
</#if>
static uint16_t BL_WriteFlash(void);
static uint16_t BL_EraseFlash(void);
<#if (SUPPORT_EE_DATA_READ == "Enabled")>
static uint16_t BL_ReadEEData(void);
</#if>
<#if (SUPPORT_EE_DATA_WRITE == "Enabled")>
static uint16_t BL_WriteEEData(void);
</#if>
static uint16_t BL_ReadConfig(void);
<#if (SUPPORT_CONFIG_WRITE == "Enabled")>
static uint16_t BL_WriteConfig(void);
</#if>
<#if VERIFY == "STATUS_BYTE_VERIFICATION">
static uint8_t BL_WriteAppStatus(void);
</#if>
static uint16_t BL_CalcChecksum(void);
static void BL_RunBootloader(void);
static bool BL_BootloadRequired(void);
static uint16_t BL_ProcessBootBuffer(void);
static void BL_CheckDeviceReset(void);


// *****************************************************************************
    static uint16_t checkSum;      // Checksum accumulator
    static bool resetPending  = false;
// Force variables into Unbanked for 1-cycle accessibility 
<#if (PIC18F_Q == "Enabled")>
    static uint8_t EE_Key_1    __at(0x500);
    static uint8_t EE_Key_2    __at(0x501);
<#else>
    static uint8_t EE_Key_1    __at(0x0);
    static uint8_t EE_Key_2    __at(0x1);
</#if>

frame_t  frame;

<#if (( DEVICE_FAMILY == "K42") || ( DEVICE_FAMILY == "K83") || (PIC18F_Q == "Enabled"))>
// *****************************************************************************
// Newer PIC18F devices have vectored interrupts with the base address specified
// by SFR IVTBASE.  Thus there is no need for the bootloader to redirect the 
// Interrupt vectors. The end application should specify IVTBASE as needed
// *****************************************************************************
<#else>
// *****************************************************************************
// The bootloader code does not use any interrupts.
// However, the application code may use interrupts.
// The interrupt vector on a PIC18F is located at
// address 0x0008 (High) and 0x0018 (Low). 
// The following function will be located
// at the interrupt vector and will contain a jump to
// the new application interrupt vectors
// *****************************************************************************

    asm ("psect  intcode,global,reloc=2,class=CODE,delta=1");
    asm ("GOTO " str(NEW_INTERRUPT_VECTOR_HIGH));

    asm ("psect  intcodelo,global,reloc=2,class=CODE,delta=1");
    asm ("GOTO " str(NEW_INTERRUPT_VECTOR_LOW));
</#if>
// *****************************************************************************
void BL_Initialize(void)
{
    EE_Key_1 = 0;
    EE_Key_2 = 0;
<#if (BOOTLOADER_RUNNING_IO == "Enabled")>
    BL_INDICATOR_OFF();
</#if>
    if ((BL_BootloadRequired() == true))
    {
        BL_INDICATOR_ON();
        BL_RunBootloader();     // generic comms layer
    }
    STKPTR = 0x00;
	BSR = 0x00;
<#if (BOOTLOADER_RUNNING_IO == "Enabled")>
	BL_INDICATOR_OFF();
</#if>
    asm ("goto  "  str(NEW_RESET_VECTOR));
}

static void BL_RunBootloader(void)
{
    volatile uint16_t  index;
    uint16_t  messageLength;
    uint8_t ch;

    while (1)
    {
        while(BL_CommunicationModuleIsReady() != true); // wait for last byte to shift out before
                                                        // starting autobaud.
        BL_CheckDeviceReset();  // Response has been sent.  Check to see if a reset was requested

        BL_CommunicationModuleInit();

        // Read and parse the data
        index = 0;       // Point to the buffer
        messageLength = 9;  // message has 9 bytes of overhead (Opcode + Length + Address)

        while (index < messageLength)
        {
            BL_CommunicationModuleRead(&ch,1);
            frame.buffer [index ++] = ch;
            if (index == 4)
            {
                if ((frame.command == WRITE_FLASH)
                 || (frame.command == WRITE_EE_DATA)
                 || (frame.command == WRITE_CONFIG))
                {
                    messageLength += frame.data_length;
                }
            }
        }

        messageLength = BL_ProcessBootBuffer ();

        // Send the data buffer back
        BL_CommunicationModuleWrite(frame.buffer, messageLength);       
    }
}

// *****************************************************************************
static bool BL_BootloadRequired(void)
{
<#if IO_PIN_ENTRY == "Enabled">
// ******************************************************************
//  Check an IO pin to force entry into bootloader
// ******************************************************************

    for (uint8_t i = 0; i != 0xFF; i++) NOP();

    if (IO_PIN_ENTRY_GetInputValue() == IO_PIN_ENTRY_RUN_BL)
    {
        return (true);
    }
</#if>

    if (BL_bootVerify() == false)
    {
        return (true);
    }

	return (false);
}

// *****************************************************************************
static uint16_t BL_ProcessBootBuffer(void)
{
    uint16_t   len;
    EE_Key_1 = frame.EE_key_1;
    EE_Key_2 = frame.EE_key_2;
	
// Test the command field and sub-command.
    switch (frame.command)
    {
    case    READ_VERSION:
        len = BL_GetVersionData();
        break;
<#if (SUPPORT_FLASH_READ == "Enabled")>
    case    READ_FLASH:
        len = BL_ReadFlash();
        break;
</#if>
    case    WRITE_FLASH:
        len = BL_WriteFlash();
        break;
    case    ERASE_FLASH:
        len = BL_EraseFlash();
        break;
<#if (SUPPORT_EE_DATA_READ == "Enabled")>
    case    READ_EE_DATA:
        len = BL_ReadEEData();
        break;
</#if>
<#if (SUPPORT_EE_DATA_WRITE == "Enabled")>
    case    WRITE_EE_DATA:
        len = BL_WriteEEData();
        break;
</#if>
    case    READ_CONFIG:
        len = BL_ReadConfig();
        break;
<#if SUPPORT_CONFIG_WRITE == "Enabled">		
    case    WRITE_CONFIG:
        len = BL_WriteConfig();
        break;
</#if>
    case    CALC_CHECKSUM:
        len = BL_CalcChecksum();
        break;
    case    RESET_DEVICE:
        frame.data[0] = COMMAND_SUCCESS;
        resetPending = true;
        len = 10;
        break;
<#if VERIFY == "STATUS_BYTE_VERIFICATION">
    case    WRITE_STATUS:
        len = BL_WriteAppStatus();
        break;
</#if>
    default:
        frame.data[0] = ERROR_INVALID_COMMAND;
        len = 10;
    }
    return (len);
}

// *****************************************************************************
// Commands
//
//        Cmd     Length----------------   Address---------------
// In:   [<0x00> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00>]
// OUT:  [<0x00> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00> <VERL><VERH>]
// *****************************************************************************
static uint16_t BL_GetVersionData(void)
{
    uint32_t configurationBytesAddress;
    configurationBytesAddress = CONFIGURATION_BYTES_START;

    frame.data[0] = MINOR_VERSION;
    frame.data[1] = MAJOR_VERSION;
    frame.data[2] = 0;       // Max packet size (256)
    frame.data[3] = 1;
    frame.data[4] = 0;
    frame.data[5] = 0;
    frame.data[6] = FLASH_Read(DEVICE_ID_LOW_ADDRESS);
    frame.data[7] = FLASH_Read(DEVICE_ID_HIGH_ADDRESS);
    frame.data[8] = 0;
    frame.data[9] = 0;

    frame.data[10] = PROGMEM_PAGE_SIZE_LOW_BYTE;
    frame.data[11] = PROGMEM_PAGE_SIZE_HIGH_BYTE;

    for (uint8_t  i= 12; i < 16; i++)
    {
        frame.data[i] = FLASH_Read(configurationBytesAddress++);
    }

    return  25;   // total length to send back 9 byte header + 16 byte payload
}

<#if (SUPPORT_FLASH_READ == "Enabled")>
// *****************************************************************************
// Read Flash
//        Cmd     Length----------------   Address---------------  Data ---------
// In:   [<0x01> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00>]
// OUT:  [<0x01> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00> <Data>..<data>]
// *****************************************************************************
static uint16_t BL_ReadFlash(void)
{
    flash_address_t address;

    address = (((flash_address_t)frame.address_U) << 16)
            | (((flash_address_t)frame.address_H) << 8)
            | (flash_address_t)frame.address_L;

    for (uint16_t i = 0; i < frame.data_length; i++)
    {
        frame.data[i] = FLASH_Read(address++);
    }

    return (frame.data_length + 9);
}
</#if>


// *****************************************************************************
// Write Flash
//        Cmd     Length----- Keys------   Address---------------  Data ---------
// In:   [<0x02> <0x00><0x00><0x55><0xAA> <0x00><0x00><0x00><0x00> <Data>..<data>]
// OUT:  [<0x02> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00> <0x01>]
// *****************************************************************************

static uint16_t BL_WriteFlash(void)
{
    flash_address_t     address;
    uint16_t            unlockKey;
    nvm_status_t        errorStatus = NVM_OK;
    flash_address_t     flashStartPageAddress;

    unlockKey = (((uint16_t)EE_Key_2) << 8)
                | (uint16_t)EE_Key_1;

    address = (((flash_address_t)frame.address_U) << 16)
            | (((flash_address_t)frame.address_H) << 8)
            |   (flash_address_t)frame.address_L;

    NVM_StatusClear();

<#if (SELF_PROTECTION == "Enabled")>
    if (address < NEW_RESET_VECTOR)
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10);
    }
</#if>

    if (frame.data_length < PROGMEM_PAGE_SIZE)
    {
        for(uint16_t i = frame.data_length; i < PROGMEM_PAGE_SIZE; i++)
        {
            frame.data[i] = 0xFF;
        }
    }

    NVM_UnlockKeySet(unlockKey);

    flashStartPageAddress = FLASH_ErasePageAddressGet(address);
    errorStatus = FLASH_RowWrite(flashStartPageAddress, frame.data);

    NVM_UnlockKeyClear();

    if (errorStatus == NVM_ERROR)
    {
        NVM_StatusClear();
    }

    frame.data[0] = errorStatus == NVM_OK ? COMMAND_SUCCESS : ERROR_INVALID_COMMAND;
    EE_Key_1 = 0x00;  // erase EE Keys
    EE_Key_2 = 0x00;
    return (10);
}

// *****************************************************************************
// Erase Program Memory
// Erases data_length rows from program memory
//        Cmd     Length----- Keys------   Address---------------  Data ---------
// In:   [<0x03> <0x00><0x00><0x55><0xAA> <0x00><0x00><0x00><0x00>]
// OUT:  [<0x03> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00> <0x01>]
// *****************************************************************************
static uint16_t BL_EraseFlash(void)
{
    flash_address_t     address;
    uint16_t            unlockKey;
    nvm_status_t        errorStatus = NVM_OK;
    flash_address_t     flashStartPageAddress;

    unlockKey = (((uint16_t)EE_Key_2) << 8)
                | (uint16_t)EE_Key_1;
    
    address = (((flash_address_t)frame.address_U) << 16)
            | (((flash_address_t)frame.address_H) << 8)
            |   (flash_address_t)frame.address_L;    

    NVM_StatusClear();

<#if (SELF_PROTECTION == "Enabled")>
    if (address < NEW_RESET_VECTOR)
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10);
    }
</#if>

    NVM_UnlockKeySet(unlockKey);

    for (uint16_t i = 0; i < frame.data_length; i++)
    {
        flashStartPageAddress = FLASH_ErasePageAddressGet(address);

        errorStatus = FLASH_PageErase(flashStartPageAddress);

        flashStartPageAddress += PROGMEM_PAGE_SIZE;

        if (errorStatus == NVM_ERROR)
        {
            NVM_StatusClear();
            break;
        }
    }

    NVM_UnlockKeyClear();

    frame.data[0]  = errorStatus == NVM_OK ? COMMAND_SUCCESS : ERROR_INVALID_COMMAND;
    frame.EE_key_1 = 0x00;  // erase EE Keys
    frame.EE_key_2 = 0x00;
    return (10);
}

<#if (SUPPORT_EE_DATA_READ == "Enabled")>
// **************************************************************************************
// Read_EE_Data
//
//        Cmd     Length-----              Address---------------  Data ---------
// In:   [<0x04> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00>]
// OUT:  [<0x04> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00> <Data>..<data>]
//
// *****************************************************************************
static uint16_t BL_ReadEEData(void)
{
    eeprom_address_t  address;

    address = (((eeprom_address_t)frame.address_U) << 16)
            | (((eeprom_address_t)frame.address_H) << 8)
            |   (eeprom_address_t)frame.address_L;

    for (uint16_t  i = 0; i < frame.data_length; i++)
    {
        frame.data[i] = EEPROM_Read(address);
        ++address;
    }

    return (frame.data_length + 9);
}
</#if>
<#if (SUPPORT_EE_DATA_WRITE == "Enabled")>
// *****************************************************************************
// Write_EE_Data
//
//        Cmd     Length-----              Address---------------  Data ---------
// In:   [<0x05> <0x00><0x00><0x55><0xAA> <0x00><0x00><0x00><0x00> <Data>..<data>]
// OUT:  [<0x05> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00> <0x01>]
// *****************************************************************************
static uint16_t BL_WriteEEData(void)
{
    eeprom_address_t    address;
    uint16_t            unlockKey;

    unlockKey = (((uint16_t)EE_Key_2) << 8)
            |     (uint16_t)EE_KEY_1;

    address = (((eeprom_address_t)frame.address_U) << 16)
            | (((eeprom_address_t)frame.address_H) << 8)
            |   (eeprom_address_t)frame.address_L;

    NVM_UnlockKeySet(unlockKey);

    for (uint16_t i = 0; i < frame.data_length; i++)
    {
        EEPROM_Write(address, frame.data[i]);
    }

    NVM_UnlockKeyClear();

    frame.data[0] = COMMAND_SUCCESS;
    return 10;
}
</#if>

// *****************************************************************************
// Read Config Words
//        Cmd     Length-----              Address---------------  Data ---------
// In:   [<0x06> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00>]
// OUT:  [<0x06> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00> <Data>..<data>]
// *****************************************************************************
static uint16_t BL_ReadConfig(void)
{
    uint32_t configurationAddress; 

    configurationAddress = (((uint32_t)frame.address_U) << 16)
                        | (((uint32_t)frame.address_H) << 8)
                        | (uint32_t)frame.address_L;


<#if (SELF_PROTECTION == "Enabled")>
    if (configurationAddress < NEW_RESET_VECTOR)
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10);
    }
</#if>
    for (uint8_t  i = 0; i < frame.data_length; i++)
    {
        frame.data[i] = EEPROM_Read(configurationAddress++);
    }
    return (9 + frame.data_length);           // 9 byte header + 4 bytes config words
}

<#if (SUPPORT_CONFIG_WRITE == "Enabled")>
// *****************************************************************************
// Write Config Words
//        Cmd     Length-----              Address---------------  Data ---------
// In:   [<0x07> <0x00><0x00><0x55><0xAA> <0x00><0x00><0x00><0x00> <Data>..<data>]
// OUT:  [<0x07> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00>]
// *****************************************************************************
static uint16_t BL_WriteConfig(void)
{
    uint32_t    configurationAddress; 
    uint16_t    unlockKey; 
    uint8_t     configurationByte = 0;

    unlockKey = (((uint16_t)EE_Key_2) << 8)
                | (uint16_t)EE_Key_1;

    configurationAddress = (((uint32_t)frame.address_U) << 16)
                        | (((uint32_t)frame.address_H) << 8)
                        | (uint32_t)frame.address_L;
<#if (SELF_PROTECTION == "Enabled")>
    if (configurationAddress < NEW_RESET_VECTOR)
    {
        frame.data[0] = ERROR_ADDRESS_OUT_OF_RANGE;
        return (10);
    }
</#if>

    NVM_UnlockKeySet(unlockKey);

    for (uint8_t i = 0; i < frame.data_length; i++)
    {

        configurationByte = EEPROM_Read(configurationAddress);
        configurationByte = frame.data[i];
        EEPROM_Write(configurationAddress, configurationByte);

        ++configurationAddress;
    }

    NVM_UnlockKeyClear();

    frame.data[0] = COMMAND_SUCCESS;
    EE_Key_1 = 0x00;  // erase EE Keys
    EE_Key_2 = 0x00;
    return (10);
}
</#if>

<#if VERIFY == "STATUS_BYTE_VERIFICATION">
// *****************************************************************************
// Write Valid Application Status
//        Cmd     Length----- Keys------   Address---------------  Data ---------
// In:   [<0x0B> <0x00><0x00><0x55><0xAA> <0x00><0x00><0x00><0x00>]
// OUT:  [<0x0B> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00> <0x01>]
// *****************************************************************************
static uint8_t BL_WriteAppStatus(void)
{
    flash_address_t address;
    flash_address_t tempAddress;
    uint16_t unlockKey;
    uint16_t dataIndex;
    nvm_status_t errorStatus = NVM_OK;

    unlockKey = (((uint16_t)EE_Key_2) << 8)
                | (uint16_t)EE_Key_1;

    address = FLASH_ErasePageAddressGet(PROGMEM_SIZE - 1);

    NVM_StatusClear();

    for (dataIndex = 0; dataIndex < (PROGMEM_PAGE_SIZE - 1); dataIndex++)
    {
        tempAddress = address;
        frame.data[dataIndex] = FLASH_Read(tempAddress++);
    }

    frame.data[dataIndex] = APPLICATION_VALID;
    dataIndex = 0;

    NVM_UnlockKeySet(unlockKey);

    errorStatus = FLASH_PageErase(address);
    if (errorStatus == NVM_OK)
    {
        errorStatus = FLASH_RowWrite(address, frame.data);
    }
    else
    {
        NVM_StatusClear();
    }

    if (errorStatus == NVM_ERROR)
    {
        NVM_StatusClear();
    }

    NVM_UnlockKeyClear();

    frame.data[dataIndex++] = errorStatus == NVM_OK ? COMMAND_SUCCESS : ERROR_INVALID_COMMAND;

    return (9 + dataIndex);
}
</#if>

// *****************************************************************************************
// Calculate Checksum
// In:	[<0x08><DataLengthL><DataLengthH> <unused><unused> <ADDRL><ADDRH><ADDRU><unused>...]
// OUT:	[9 byte header + ChecksumL + ChecksumH]
//        Cmd     Length-----              Address---------------  Data ---------
// In:   [<0x08> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00>]
// OUT:  [<0x08> <0x00><0x00><0x00><0x00> <0x00><0x00><0x00><0x00> <CheckSumL><CheckSumH>]
// *****************************************************************************
static uint16_t BL_CalcChecksum(void)
{
    flash_address_t address;
#if PROGMEM_SIZE > 0x10000
    uint32_t  i;
    uint32_t  length;
#else
    uint16_t  i;
    uint16_t  length;
#endif 
    address = (((flash_address_t)frame.address_U) << 16)
            | (((flash_address_t)frame.address_H) << 8)
            | (flash_address_t)frame.address_L;

    checkSum = 0;
    length = frame.data_length;
#if PROGMEM_SIZE > 0x10000
    length += ((uint32_t) frame.EE_key_1) << 16;
#endif 
    for (i = 0; i < length; i += 2)
    {
        checkSum += (uint16_t)FLASH_Read(address++);
        checkSum += ((uint16_t)FLASH_Read(address++)) << 8;
     }
     frame.data[0] = (uint8_t) (checkSum & 0x00FF);
     frame.data[1] = (uint8_t)((checkSum & 0xFF00) >> 8);
     return (11);
}

// *****************************************************************************
// Check to see if a device reset had been requested.  We can't just reset when
// the reset command is issued.  Instead we have to wait until the acknowledgement
// is finished sending back to the host.  Then we reset the device.
// *****************************************************************************
static void BL_CheckDeviceReset(void)
{
    if (resetPending == true)
    {
<#if (VERIFY == "STATUS_BYTE_VERIFICATION")>
        // *************************************
        // Write the status of the bootloader at the device reset
        // which should occur after every reprogramming action
        // *************************************
        if(FLASH_Read(STATUS_ADDRESS) == 0xFF)
        {   
            uint16_t        unlockKey;
            nvm_status_t    updateStatus;

            unlockKey = (((uint16_t)EE_Key_2) << 8)
                        | (uint16_t)EE_Key_1;

            NVM_StatusClear();

            NVM_UnlockKeySet(unlockKey);

            updateStatus = FLASH_Write(STATUS_ADDRESS, (uint16_t)APPLICATION_VALID);

            NVM_UnlockKeyClear();

            if (updateStatus == NVM_ERROR)
            {
                NVM_StatusClear();
            }
        }
</#if>
<#if (BOOTLOADER_RUNNING_IO == "Enabled")>
        BL_INDICATOR_OFF();
</#if>
        RESET();
    }
    return;
}
