${disclaimer}

//
//
//
// Memory Map
//   -----------------
//   |    0x0000     |   Reset vector
//   |               |
//   |    0x0004     |   Interrupt vector
//   |               |
//   |               |
//   |  Boot Block   |   (this program)
//   |               |
//   |    0x0300     |   Re-mapped Reset Vector
//   |    0x0304     |   Re-mapped High Priority Interrupt Vector
//   |               |
//   |       |       |
//   |               |
//   |  Code Space   |   User program space
//   |               |
//   |       |       |
//   |               |
//   |    0x3FFF     |
//   -----------------
//
//
//
// Definitions:
//
//   STX     -   Start of packet indicator
//   DATA    -   General data up to 255 bytes
//   COMMAND -   Base command
//   DLEN    -   Length of data associated to the command
//   ADDR    -   Address up to 24 bits
//   DATA    -   Data (if any)
//
//
// Commands:
//
//   RD_VER      0x00    Read Version Information
//   RD_MEM      0x01    Read Program Memory
//   WR_MEM      0x02    Write Program Memory
//   ER_MEM      0x03    Erase Program Memory (NOT supported by PIC16)
//   RD_EE       0x04    Read EEDATA Memory
//   WR_EE       0x05    Write EEDATA Memory
//   RD_CONFIG   0x06    Read Config Memory (NOT supported by PIC16)
//   WT_CONFIG   0x07    Write Config Memory (NOT supported by PIC16)
//   CHECKSUM    0x08    Calculate 16 bit checksum of specified region of memory
//   RESET       0x09    Reset Device and run application
//
// *****************************************************************************




// *****************************************************************************
    #include <stdbool.h>
    #include "../bl_bootload.h"
    #include"no_comms_BL.X/mcc_generated_files/EEPROM3_driver.h"


// *****************************************************************************

// *****************************************************************************
#define	MINOR_VERSION   0x00       // Version
#define	MAJOR_VERSION   0x00

#define ERROR_ADDRESS_OUT_OF_RANGE   0xFE
#define ERROR_INVALID_COMMAND        0xFF
#define COMMAND_SUCCESS              0x01

// To be device independent, these are set by mcc in memory.h
#define  LAST_WORD_MASK          (WRITE_FLASH_BLOCKSIZE - 1)
#define  NEW_RESET_VECTOR        ${NEW_RESET_VECTOR_U}
#define  NEW_INTERRUPT_VECTOR    ${NEW_INTERRUPT_VECTOR_LOW_U}

#define  SOURCE_BOOTBLOCK_ADDRESS  0x0000UL
#define _str(x)  #x
#define str(x)  _str(x)

// *****************************************************************************
bool Compare_Images (uint16_t  Dest_Address, uint16_t  Source_Address, uint16_t Length);
void Copy_Image (uint16_t  Dest_Address, uint16_t Source_Address, uint16_t Length);
void Source_Read (uint8_t Buffer [], uint16_t  Start_Address, uint16_t Length);
void Source_Write (uint8_t Buffer [], uint16_t Start_Address, uint16_t Length);
void Read_Flash (uint8_t Buffer [], uint16_t  Start_Address, uint16_t Length);
uint8_t  Write_Flash (uint16_t  Start_Address,  uint8_t Buffer [], uint16_t Length);
uint8_t Erase_Flash (uint16_t  Start_Address, uint16_t Length);
void StartWrite(void);


// *****************************************************************************
    BootBlock_t  BootBlock;

// Force variables into Unbanked for 1-cycle accessibility 
    uint8_t EE_Key_1    @ 0x70 = 0;
    uint8_t EE_Key_2    @ 0x71 = 0;



// *****************************************************************************
// The bootloader code does not use any interrupts.
// However, the application code may use interrupts.
// The interrupt vector on a PIC16F is located at
// address 0x0004. 
// The following function will be located
// at the interrupt vector and will contain a jump to
// the new application interrupt vector
#asm
    psect  intentry delta=2
    pagesel NEW_INTERRUPT_VECTOR
    GOTO    NEW_INTERRUPT_VECTOR
#endasm

void BL_Initialize ()
{
<#if (Bootloader_Running_IO == "Enabled")>
	BL_INDICATOR_ON();
</#if>
    Source_Read ((uint8_t *)&BootBlock, SOURCE_BOOTBLOCK_ADDRESS, sizeof (BootBlock));
    EE_Key_1 = BootBlock.EE_Key_1;
    EE_Key_2 = BootBlock.EE_Key_2;
    
    if (BootBlock.Trigger == RUN_BOOTLOADER)
    {
        Copy_Image (BootBlock.Destination_Address.Address, BootBlock.Source_Address.Address, BootBlock.Length);
    }
    STKPTR = 0x1F;
<#if (Bootloader_Running_IO == "Enabled")>
    BL_INDICATOR_OFF();
</#if>
    asm ("pagesel " str(NEW_RESET_VECTOR));
    asm ("goto  "  str(NEW_RESET_VECTOR));
}


// **************************************************************************************
// Source Read
void Source_Read (uint8_t Buffer [], uint16_t Start_Address, uint16_t Length)
{
    // Insert call to your device read function here
    EEPROM3_ReadBlock((uint32_t) Start_Address, Buffer, Length);
}

// **************************************************************************************
// Source Write
void Source_Write (uint8_t Buffer [], uint16_t Start_Address, uint16_t Length)
{
    // Insert call to your device write function here
    EEPROM3_WriteBlock ((uint32_t)Start_Address, Buffer, Length);
}


bool Compare_Images (uint16_t  Dest_Address, uint16_t  Source_Address, uint16_t Length)
{
    bool     Image_Valid;
    uint16_t  dest;
    uint16_t  offset;
    address_t  Address;
    uint8_t  Source_Buffer [WRITE_FLASH_BLOCKSIZE*2];

    Image_Valid = true;
    for (offset = 0; offset < Length/2; offset += WRITE_FLASH_BLOCKSIZE)
    {
        Source_Read ((uint8_t *)Source_Buffer, (Source_Address + 2 * offset), sizeof (Source_Buffer));
        for (dest = 0; dest < WRITE_FLASH_BLOCKSIZE; dest ++)
        {
            Address.Address = Dest_Address + offset + dest;
            ${EEADRL_SFR} = Address.LowByte;
            ${EEADRH_SFR} = Address.HighByte;
            ${EECON1_SFR}bits.RD = 1;
            if ((${EEDATL_SFR} != Source_Buffer [2*dest])
             || (${EEDATH_SFR} != Source_Buffer [2*dest+1]))
            {
                Image_Valid = false;
                break;  // no need to continue
            }
        }
    }
    return Image_Valid;
}

void Copy_Image (uint16_t Dest_Address, uint16_t Source_Address, uint16_t Length)
{
    uint8_t   Copy_Buffer [WRITE_FLASH_BLOCKSIZE * 2];
    uint16_t  dest;
    uint16_t  src;

    STKPTR = 0x1F;  // reset stack pointer (we exit this function via a reset and won't need the calls on the stack).
    Erase_Flash (Dest_Address, (Length/ERASE_FLASH_BLOCKSIZE));

    dest = Dest_Address;
    for (src = Source_Address; src < Source_Address + Length; src += WRITE_FLASH_BLOCKSIZE * 2)
    {
        Source_Read (Copy_Buffer, src, sizeof (Copy_Buffer));
        Write_Flash (dest, Copy_Buffer, sizeof (Copy_Buffer));
        dest += WRITE_FLASH_BLOCKSIZE;
    }

    if (Compare_Images (Dest_Address, Source_Address, Length) == true)
    {
        ${EEDATL_SFR} = 0;
        Copy_Buffer [0] = 0x00;
        Copy_Buffer [1] = 0x00;
        Copy_Buffer [2] = 0x00;
        Source_Write (Copy_Buffer, SOURCE_BOOTBLOCK_ADDRESS+10, 3);
        EE_Key_1 = 0x00;
        EE_Key_2 = 0x00;
    }
<#if (Bootloader_Running_IO == "Enabled")>
    BL_INDICATOR_OFF();
</#if>
    RESET();
}



// **************************************************************************************
// Read Flash
void Read_Flash (uint8_t Buffer [], uint16_t  Start_Address, uint16_t Length)
{
    address_t  Address;
    
    Address.Address = Start_Address;
    ${EEADRL_SFR} = Address.LowByte;
    ${EEADRH_SFR} = Address.HighByte;
    ${EECON1_SFR} = 0x00;     
    for (uint8_t  i= 0; i < Length; i += 2)
    {
        ${EECON1_SFR}bits.RD = 1;
        Buffer[i]   = ${EEDATL_SFR};
        Buffer[i+1] = ${EEDATH_SFR};
        if (++ ${EEADRL_SFR} == 0x00)
            ++ ${EEADRH_SFR};
    }
}


// **************************************************************************************
// Write Flash
// In:	[<0x02><DLENBLOCK><0x55><0xAA><ADDRL><ADDRH><ADDRU><DATA>...]
// OUT:	[<0x02>]
uint8_t  Write_Flash (uint16_t  Start_Address,  uint8_t Buffer [], uint16_t Length)
{
    address_t  Address;
    
    if (Start_Address < NEW_RESET_VECTOR)
    {
        return (ERROR_ADDRESS_OUT_OF_RANGE);
    }
    Address.Address = Start_Address;
    ${EEADRL_SFR} = Address.LowByte;
    ${EEADRH_SFR} = Address.HighByte;
    ${EECON1_SFR} = 0xA4;       // Setup writes
    for (uint8_t  i= 0; i < Length; i += 2)
    {
        if (((${EEADRL_SFR} & LAST_WORD_MASK) == LAST_WORD_MASK)
          || (i == Length - 2))
            ${EECON1_SFR}bits.LWLO = 0;
        ${EEDATL_SFR} = Buffer[i];
        ${EEDATH_SFR} = Buffer[i+1];

        StartWrite();
        if (++ ${EEADRL_SFR} == 0x00)
            ++ ${EEADRH_SFR};
    }
    return (COMMAND_SUCCESS);
}

// **************************************************************************************
// Erase Program Memory
// Erases data_length rows from program memory
uint8_t Erase_Flash (uint16_t  Start_Address, uint16_t Length)
{
    address_t  Address;
    
    if (Start_Address < NEW_RESET_VECTOR)
    {
        return (ERROR_ADDRESS_OUT_OF_RANGE);
    }
    Address.Address = Start_Address;
    ${EEADRL_SFR} = Address.LowByte;
    ${EEADRH_SFR} = Address.HighByte;
    for (uint16_t i=0; i < Length; i += 2)  // length is in bytes
    { 
        if ((${EEADRH_SFR} & 0x7F) >= ((PROGMEM_SIZE & 0xFF00)>>8))
        {
            return (ERROR_ADDRESS_OUT_OF_RANGE);
        }
        ${EECON1_SFR} = 0x94;       // Setup writes
        StartWrite();
        if ((${EEADRL_SFR} += ERASE_FLASH_BLOCKSIZE) == 0x00)
            ++ ${EEADRH_SFR};
    }
    return (10);
}




// *****************************************************************************
// Unlock and start the write or erase sequence.

void StartWrite()
{
    CLRWDT();
//    EECON2 = EE_Key_1;
//    EECON2 = EE_Key_2;
//    EECON1bits.WR = 1;       // Start the write
// had to switch to assembly - compiler doesn't comprehend no need for bank switch
    asm ("movf " str(_EE_Key_1) ",w");
    asm ("movwf " str(BANKMASK(${EECON2_SFR})));
    asm ("movf  " str(_EE_Key_2) ",w");
    asm ("movwf " str(BANKMASK(${EECON2_SFR})));
    asm ("bsf  "  str(BANKMASK(${EECON1_SFR})) ",1");       // Start the write
    return;
}

// *****************************************************************************
// *****************************************************************************
