import { CFunction } from "@microchip/scf-interface";

/**
 * @type A special type that is used to interface with the device_meta processed payload
 */
export type Architecture =
    | "PIC10"
    | "PIC12"
    | "PIC16"
    | "PIC18"
    | "AVR8"
    | "AVR8X"
    | "AVR" // bootloader specific value
    | "TINY" // bootloader specific value
    | "PIC24"
    | "dsPIC33"
    | "PIC32"
    | undefined;

/**
 * @interface [BootTemplatePayload] the iterface that will be used to pass the data values to the templates
 */
export interface BootTemplatePayload {
    BL_UARTWrite?: string;
    BL_UARTRead?: string;
    BL_UARTIsRxReady?: string;
    BL_UARTIsTxReady?: string;
    BL_UARTTransmitDone?: string;
    BL_UARTAutoBaudSet?: string;
    BL_UARTAutoBaudQuery?: string;
    BL_USARTIsAutoBaudDetectError?: string;
    BL_USARTAutoBaudDetectErrorReset?: string;
    DELAY_HEADER_PATH?: string;
    NVM_ACCESS_TYPE?: NVM_ACCESS_VALUE;
    RESET_BIT_U?: RESET_BITMASK_VALUE | FTL_NULL_VALUE;
    FUSE?: AVR_FUSE_VALUE | FTL_NULL_VALUE;
    SUPPORT_FLASH_READ?: FTL_BOOLEAN_VALUE;
    SUPPORT_EE_DATA_READ?: FTL_BOOLEAN_VALUE;
    SUPPORT_EE_DATA_WRITE?: FTL_BOOLEAN_VALUE;
    ERASE_CONTROL?: ERASE_CONTROL_VALUE;
    AVR_PAGE?: MEMORY_SIZE_VALUE | FTL_NULL_VALUE;
    NEW_RESET_VECTOR_U?: string;
    NEW_INTERRUPT_VECTOR_LOW_U?: string;
    NEW_INTERRUPT_VECTOR_HIGH_U?: string;
    SELF_PROTECTION?: FTL_BOOLEAN_VALUE;
    CFGS_BIT?: "NVMREGS" | "EECON1";
    EECON1_SFR?: "NVMCON1" | "EECON1";
    EECON2_SFR?: "NVMLOCK" | "NVMCON2";
    EEDATL_SFR?: "NVMDATL" | "EEDATL";
    EEDATH_SFR?: "NVMDATH" | "EEDATH";
    EEADRL_SFR?: "NVMADRL" | "EEARDL";
    EEADRH_SFR?: "NVMADRH" | "EEARDH";
    DRAIN_RXBUF_AFTER_AB?: FTL_BOOLEAN_VALUE | FTL_NULL_VALUE;
    SUPPORT_CONFIG_WRITE?: FTL_BOOLEAN_VALUE;
    SUPPORT_CONFIG_READ?: FTL_BOOLEAN_VALUE;
    SUPPORT_CHECKSUM?: FTL_BOOLEAN_VALUE;
    PIC18F_Q?: string;
    DEVICE_FAMILY?: string;
    DEVICE_TYPE?: DEVICE_TYPE_VALUES | FTL_NULL_VALUE;
    CHIP_TYPE?: Architecture;
    APP_VALID_VALUE_U?: "0x55";
    NO_APPLICATION_CODE_U?: EMPTY_DATA_VALUE;
    CRC_SEED_U?: CRC_SEED_VALUE | FTL_NULL_VALUE;
    CRC_SIZE_U?: CRC_SIZE_VALUE | FTL_NULL_VALUE;
    CRC_POLYNOMIAL_U?: CRC_POLYNOMIAL_VALUE | FTL_NULL_VALUE;
    CRC_XOR_U?: "0x00000000";
    CHECKSUM_SIZE_U?: CRC_SIZE_VALUE | FTL_NULL_VALUE;
    VERIFY?: VERIFICATION_SCHEME_VALUE | FTL_NULL_VALUE;
    BOOTLOADER_RUNNING_IO?: FTL_BOOLEAN_VALUE;
    IO_PIN_ENTRY?: FTL_BOOLEAN_VALUE;
    BL_INDICATOR_PIN_ACTIVATION?: ACTIVATION_STRING_VALUE;
    BL_ENTRY_POINT?: string;
    BL_INDICATOR_PIN?: string;
    DEVICE_ID_START_ADDRESS?: string;
    FLASH_SIZE_U?: string;
    EEPROM_START_U?: string;
    CONFIGURATIONS_BYTES_START_ADDRESS_U?: string;
    USER_ID_START_ADDRESS_U?: string;
    BOOT_BLOCK_SIZE_U?: BOOT_BLOCK_SIZE_VALUE;
    EEPROM_SIZE_U?: string;
}

/**
 * @interface [BootUserData] interface to warrent configuration of the injector class. Prevents sending unexpected data from derived data.
 */
export interface BootUserData {
    BootRunningIO: boolean | undefined;
    BootEntryIO: boolean | undefined;
    FlashRead: boolean | undefined;
    EERead: boolean | undefined;
    EEWrite: boolean | undefined;
    ConfigWrite: boolean | undefined;
    ConfigRead: boolean | undefined;
    EraseMode: ERASE_CONTROL_VALUE | undefined;
    verification: string | undefined | VERIFICATION_SCHEME_VALUE;
    ResetVectorAddress: number;
    SelfProtect: boolean | undefined;
    IndicatorData: string | undefined;
    EntryData: string | undefined;
    Checksum: boolean | undefined;
}

// Bootloader Acceptable Template Data Values
//      With this approach we can set the values that our templates are allowed to have and then we can also pass these types between the different functions
//      and the deriveddata class. This approach does not change how the derived data class interacts with MCC but instead it forced correctness
//      on the prgrammers of this project to use pre-defined values in all functions that push out template data to the derived data class
export type FTL_NULL_VALUE = "NotApplicable";
export type EMPTY_DATA_VALUE = "0xFF" | "0x3FFF";
export type CRC_SEED_VALUE = "0xFFFF" | "0xFFFFFFFF";
export type CRC_SIZE_VALUE = "2" | "4";
export type CRC_POLYNOMIAL_VALUE = "0x1021" | "0xEDB88320";
export type FTL_BOOLEAN_VALUE = "Enabled" | "Disabled";
export type NVM_ACCESS_VALUE = "ROW_ACCESS" | "BYTE_ACCESS";
export type AVR_FUSE_VALUE = "Reserved_Two_Byte" | "Reserved_One_Byte";
export type VERIFICATION_SCHEME_VALUE =
    | "RESET_VECTOR_VERIFICATION"
    | "STATUS_BYTE_VERIFICATION"
    | "CRC16_VERIFICATION"
    | "CRC32_VERIFICATION"
    | "CHECKSUM_VERIFICATION";
export type RESET_BITMASK_VALUE = "RSTCTRL_SWRST_bm" | "RSTCTRL_SWRE_bm";
export type ERASE_CONTROL_VALUE = "FullErase" | "BlockBased";
export type MEMORY_SIZE_VALUE = "Expanded" | "Classic";
export type BOOT_BLOCK_SIZE_VALUE = "512bytes" | "256bytes";
export type DEVICE_TYPE_VALUES = "AVR" | "PIC";
export type ACTIVATION_STRING_VALUE = "Active_High" | "Active_Low";

export type UART_API = {
    moduleName: string | undefined;
    Write: CFunction | "unimplemented";
    Read: CFunction | "unimplemented";
    isRxReady: CFunction | "unimplemented";
    isTxReady: CFunction | "unimplemented";
    TransmitDone: CFunction | "unimplemented";
    AutoBaudSet: CFunction | "unimplemented";
    AutoBaudQuery: CFunction | "unimplemented";
    AutoBaudDetectError: string | "unimplemented";
    AutoBaudReset: string | "unimplemented";
};

export type BOOT_MEMORY_PARAMETERS = {
    eeprom_size: number;
    eeprom_start: number;
    flash_size: number;
};

export type BOOT_COMMUNICATION_PROTOCOLS = {
    UART: UART_API | undefined;
    I2C: undefined;
};
