import {
    EmptyDerivedData,
    getModel,
    mockState,
} from "../generated_module/tests/AppModel.test";
import {
    mockAssignedModule,
    mockUnassignedModule,
} from "../generated_module/tests/processor.test";
import { MyDerivedData, getDerivedData } from "./DerivedData";
import * as Processor from "@microchip/scf-common/lib/Processor";
import { DerivedData } from "@microchip/melody-automodule-interface";
import { AppModel } from "../generated_module/src/types/AutoModuleTypes";
import { Architecture } from "./BootTemplateInterfaces";
import { getCreatorFunctions } from "./CreatorFunctions";

// Test Helpers
const initModel = (
    state: any,
): {
    model: any;
    derivedData: any;
    derivedFunctions: any;
} => {
    const newModel = getModel(mockAssignedModule(state), getTestDerivedData);
    return {
        model: newModel,
        derivedData: newModel?.getDerivedData(),
        derivedFunctions: newModel?.getDerivedData()?.getMyFunctions(),
    };
};

const getTestDerivedData = (dataModel: AppModel): DerivedData => {
    if (dataModel) {
        return new TestDerivedData(dataModel);
    }
    return new EmptyDerivedData();
};

class TestDerivedData extends MyDerivedData {
    constructor(dataModel: AppModel) {
        super(dataModel); // call the inherited constructor
    }

    public setEEPROM_Size = (newSize: number): void => {
        this.eeprom_size = newSize;
    };

    public getEEPROM_Size = (): number => {
        return this.eeprom_size;
    };

    public setDeviceArch = (arch: Architecture): void => {
        this.deviceArch = arch;
    };

    public getDeviceArch = (): Architecture => {
        return this.deviceArch;
    };
}

// Default Models
const BootloaderDefaultModel = initModel(mockState(false));

describe("@microchip/bootloader-8bit", () => {
    describe("General Derived Data Tests", () => {
        it("Should cover the empty case", () => {
            const app: any = undefined;
            const creatorTestAPI = getCreatorFunctions();

            expect(creatorTestAPI?.shouldImport?.("I2C", undefined)).toBeFalsy();

            expect(getDerivedData(app).getMyFunctions()).toEqual({});
            expect(getDerivedData(app).getModel()).toBeUndefined();
        });

        it("Test hex validator function", () => {
            expect(
                BootloaderDefaultModel?.derivedFunctions?.bootEndAddressValidator(),
            ).toEqual({
                minimum: 0x0,
                maximum: 0x20000,
            });
        });
        it("Should be able to update state data using test helper setters/getters", () => {
            const testValue = 100;
            BootloaderDefaultModel.derivedData?.setEEPROM_Size(testValue);
            expect(BootloaderDefaultModel.derivedData?.getEEPROM_Size()).toBe(testValue);
        });
    });

    describe("Import Interface Tests", () => {
        describe("DerivedData system_init_import interface test cases -- >", () => {
            const sysInitImport = {
                initializer: {
                    name: "BL_Initialize",
                    header: { name: "bl_bootload.h", path: "bootloader/" },
                },
            };
            it("Test for system_init_import is name and version defined", () => {
                expect(
                    BootloaderDefaultModel.model?.getImports()?.initializer_system
                        ?.interfaceId.name,
                ).toEqual("initializer-system");
                expect(
                    BootloaderDefaultModel.model?.getImports()?.initializer_system
                        ?.interfaceId.version,
                ).toEqual("^0.5.3");
            });
            it("Test for system_init_import value", () => {
                expect(
                    BootloaderDefaultModel.derivedFunctions?.initializer_system_results()
                        ?.initializer?.name,
                ).toEqual(sysInitImport?.initializer?.name);
            });
        });
        describe("DerivedData Delay interface test cases -- >", () => {
            it("Test for delay_blocking is name and version defined", () => {
                expect(
                    BootloaderDefaultModel.model?.getImports()?.delay_blocking
                        ?.interfaceId.name,
                ).toEqual("delay-blocking");
                expect(
                    BootloaderDefaultModel.model?.getImports()?.delay_blocking
                        ?.interfaceId.version,
                ).toEqual("^0.1.1");
            });
        });
        describe("DerivedData Device_Meta interface test cases -- >", () => {
            it("Test for device_meta is name and version defined", () => {
                expect(
                    BootloaderDefaultModel.model?.getImports()?.device_meta?.interfaceId
                        .name,
                ).toEqual("device-meta");
                expect(
                    BootloaderDefaultModel.model?.getImports()?.device_meta?.interfaceId
                        .version,
                ).toEqual("^1.*");
            });
        });
        describe("DerivedData Memory-Parameters interface test cases -- >", () => {
            it("Test for memory_parameters is name and version defined", () => {
                expect(
                    BootloaderDefaultModel.model?.getImports()?.memory_parameters
                        ?.interfaceId.name,
                ).toEqual("memory-parameters");
                expect(
                    BootloaderDefaultModel.model?.getImports()?.memory_parameters
                        ?.interfaceId.version,
                ).toEqual("1.3.0");
            });
        });
        describe("DerivedData Pin-Standard interface test cases -- >", () => {
            it("Test for pin_standard is name and version defined", () => {
                expect(
                    BootloaderDefaultModel.model?.getImports()?.pin_standard?.interfaceId
                        .name,
                ).toEqual("pin-standard");
                expect(
                    BootloaderDefaultModel.model?.getImports()?.pin_standard?.interfaceId
                        .version,
                ).toEqual("^0.*");
            });
        });
        describe("DerivedData Project-Properties interface test cases -- >", () => {
            it("Test for project_properties is name and version defined", () => {
                expect(
                    BootloaderDefaultModel.model?.getImports()?.project_properties
                        ?.interfaceId.name,
                ).toEqual("project-properties");
                expect(
                    BootloaderDefaultModel.model?.getImports()?.project_properties
                        ?.interfaceId.version,
                ).toEqual("1.0.1");
            });
        });
        describe("DerivedData Config-Request interface test cases -- >", () => {
            it("Test for config_request is name and version defined", () => {
                expect(
                    BootloaderDefaultModel.model?.getImports()?.config_request
                        ?.interfaceId.name,
                ).toEqual("config-request");
                expect(
                    BootloaderDefaultModel.model?.getImports()?.config_request
                        ?.interfaceId.version,
                ).toEqual("2.0.2");

                // testing argument load
                expect(
                    BootloaderDefaultModel.derivedFunctions?.config_request_args?.(),
                ).toEqual([
                    "bootsizeBootsize",
                    "codesizeCodesize",
                    "bootendBootend",
                    "appendAppend",
                ]);
            });
        });
        describe("DerivedData Uart-Basic interface test cases -- >", () => {
            it("Test for uart_basic is name and version defined", () => {
                expect(
                    BootloaderDefaultModel.model?.getImports()?.uart_basic?.interfaceId
                        .name,
                ).toEqual("uart-basic-drv-interface");
                expect(
                    BootloaderDefaultModel.model?.getImports()?.uart_basic?.interfaceId
                        .version,
                ).toEqual("2.1.0");
            });
        });
    });

    describe("Alert Tests", () => {
        let localDefaultModel = initModel(mockState(false));
        const nullModel =
            getModel(mockUnassignedModule(), getTestDerivedData) ?? undefined;

        it("Should return alerts for pin activations", () => {
            localDefaultModel = initModel({
                ...mockState(false),
                ...{
                    main: {
                        memory: { bootEndAddress: 0 },
                        pin: { ioPinIndicator: true, ioPinEntry: true },
                        command: {
                            configRead: true,
                            configWrite: false,
                            flashRead: true,
                        },
                    },
                },
            });
            expect(localDefaultModel?.derivedFunctions?.alerts?.()).toEqual([
                {
                    text: "Connection to PLIB was successful. Auto Baud has been enabled by default",
                    type: "hint",
                },
                {
                    text: "Allowing the bootloader to read the flash space is seen as a security risk in most systems. Consider the impacts of this command on your system.",
                    type: "hint",
                },
                {
                    text: "Select a pin for IO Pin Entry in the Pin Manager: Grid View. 'Pin Grid View' -> BOOT:ENTRY",
                    type: "error",
                },
                {
                    text: "Select a pin for IO Pin Indicator in the Pin Manager: Grid View. 'Pin Grid View' -> BOOT:INDICATE",
                    type: "error",
                },
            ]);

            localDefaultModel?.derivedData.setDeviceArch("AVR");

            expect(localDefaultModel?.derivedFunctions?.alerts?.()).toEqual([
                {
                    text: "Connection to PLIB was successful. Auto Baud has been enabled by default",
                    type: "hint",
                },
                {
                    text: "Allowing the bootloader to read the flash space is seen as a security risk in most systems. Consider the impacts of this command on your system.",
                    type: "hint",
                },
                {
                    text: "Select a pin for IO Pin Entry in the Pin Manager: Grid View. 'Pin Grid View' -> BOOT:ENTRY",
                    type: "error",
                },
                {
                    text: "Select a pin for IO Pin Indicator in the Pin Manager: Grid View. 'Pin Grid View' -> BOOT:INDICATE",
                    type: "error",
                },
                {
                    text: "The bootloader uses the device id by default. Please enable the 'Generate SIGROW APIs' option located in the NVM driver tab",
                    type: "warning",
                },
                {
                    text: "Recommended BOOTEND Range: Minimum Size - Maximum Size\n\nOptimization 0 -> 0x1C - 0x2D\nOptimization 2 -> 0x0B - 0x0D",
                    type: "warning",
                },
            ]);

            localDefaultModel?.derivedData.setDeviceArch("PIC16");

            expect(localDefaultModel?.derivedFunctions?.alerts?.()).toEqual([
                {
                    text: "Connection to PLIB was successful. Auto Baud has been enabled by default",
                    type: "hint",
                },
                {
                    text: "Allowing the bootloader to read the flash space is seen as a security risk in most systems. Consider the impacts of this command on your system.",
                    type: "hint",
                },
                {
                    text: "Select a pin for IO Pin Entry in the Pin Manager: Grid View. 'Pin Grid View' -> BOOT:ENTRY",
                    type: "error",
                },
                {
                    text: "Select a pin for IO Pin Indicator in the Pin Manager: Grid View. 'Pin Grid View' -> BOOT:INDICATE",
                    type: "error",
                },
                {
                    text: "The bootloader uses the device id and configuration area APIs by default. Please enable the 'Generate Device Id APIs' and the 'Generate Configuration Settings APIs' options located in the NVM driver tab",
                    type: "warning",
                },
                {
                    text: "Boot End Address cannot be set to 0x0",
                    type: "error",
                },
            ]);

            expect(nullModel?.getDerivedFunctions()?.alerts?.()).toEqual([
                {
                    type: "error",
                    text: "Bootloader requires a connection to the UART PLIB. Select a UART version from the 'UART(None)' tab.",
                },
            ]);
        });
    });

    describe("Hide UI EEPROM Component", () => {
        it("Should not return hidden in the behavior if the device has a valid eeprom", () => {
            expect(
                BootloaderDefaultModel.derivedFunctions?.isVisible?.("eeDataRead"),
            ).toBe(true);
        });

        it("Should return a hidden ui behavior payload when EEPROM data is invalid", () => {
            BootloaderDefaultModel.derivedData?.setEEPROM_Size(0); // Using test helpers to force eeprom_size to be 0
            expect(
                BootloaderDefaultModel.derivedFunctions?.isVisible?.("eeDataRead"),
            ).toBe(false);
        });
    });

    describe("Help UI/URL Tests", () => {
        describe("Reset Vector Verification", () => {
            it("Should return the help links and ui description", () => {
                const localDefaultModel = initModel(mockState(false));
                // Default config is reset vector; just test for its returns
                expect(
                    localDefaultModel?.derivedFunctions?.getUiBehavior?.(
                        "verificationSelector",
                    ),
                ).toEqual({
                    "ui:help":
                        "This scheme checks the address after the reset vector for a non-null byte value",
                });

                // Check the URL
                expect(
                    localDefaultModel?.derivedFunctions?.getHelpUrl?.(
                        "verificationSelector",
                        "v2/keyword-lookup?keyword=MELODY.BOOTLOADER8BIT.UI.${verify};&redirect=true",
                    ),
                ).toEqual(
                    "v2/keyword-lookup?keyword=MELODY.BOOTLOADER8BIT.UI.Check.Reset.Vector;&redirect=true",
                );
            });
        });
        describe("Status Byte Verification", () => {
            // Set Check State Flag Verification in the current state
            const localDefaultModel = initModel({
                ...mockState(false),
                ...{ main: { verify: { verificationSelector: "Check State Flag" } } },
            });
            it("Should return the help links and ui description", () => {
                expect(
                    localDefaultModel?.derivedFunctions?.getUiBehavior?.(
                        "verificationSelector",
                    ),
                ).toEqual({
                    "ui:help":
                        "This scheme will use a status byte to notify the bootloader of a success or a failure",
                });
                expect(
                    localDefaultModel?.derivedFunctions?.getHelpUrl?.(
                        "verificationSelector",
                        "v2/keyword-lookup?keyword=MELODY.BOOTLOADER8BIT.UI.${verify};&redirect=true",
                    ),
                ).toEqual(
                    "v2/keyword-lookup?keyword=MELODY.BOOTLOADER8BIT.UI.Check.State.Flag;&redirect=true",
                );
            });
        });

        describe("CRC-32 Verification", () => {
            // Setup for CRC32
            const localDefaultModel = initModel({
                ...mockState(false),
                ...{ main: { verify: { verificationSelector: "CRC-32" } } },
            });
            it("Should return the help links and ui description", () => {
                expect(
                    localDefaultModel?.derivedFunctions?.getUiBehavior?.(
                        "verificationSelector",
                    ),
                ).toEqual({
                    "ui:help":
                        "This scheme uses a Cyclic Redundancy Check algorithm with a 32-bit result to validate the application image.",
                });
                expect(
                    localDefaultModel?.derivedFunctions?.getHelpUrl?.(
                        "verificationSelector",
                        "v2/keyword-lookup?keyword=MELODY.BOOTLOADER8BIT.UI.${verify};&redirect=true",
                    ),
                ).toEqual(
                    "v2/keyword-lookup?keyword=MELODY.BOOTLOADER8BIT.UI.CRC32;&redirect=true",
                );
            });
        });
        describe("CRC-16 Verification", () => {
            // Set CRC16
            const localDefaultModel = initModel({
                ...mockState(false),
                ...{ main: { verify: { verificationSelector: "CRC-16" } } },
            });
            it("Should return the help links and ui description", () => {
                expect(
                    localDefaultModel?.derivedFunctions?.getUiBehavior?.(
                        "verificationSelector",
                    ),
                ).toEqual({
                    "ui:help":
                        "This scheme uses a Cyclic Redundancy Check algorithm with a 16-bit result to validate the application image.",
                });
                expect(
                    localDefaultModel?.derivedFunctions?.getHelpUrl?.(
                        "verificationSelector",
                        "v2/keyword-lookup?keyword=MELODY.BOOTLOADER8BIT.UI.${verify};&redirect=true",
                    ),
                ).toEqual(
                    "v2/keyword-lookup?keyword=MELODY.BOOTLOADER8BIT.UI.CRC16;&redirect=true",
                );
            });
        });
        describe("Checksum Verification", () => {
            // Set Checksum
            const localDefaultModel = initModel({
                ...mockState(false),
                ...{ main: { verify: { verificationSelector: "Checksum" } } },
            });
            it("Should return the help links and ui description", () => {
                expect(
                    localDefaultModel?.derivedFunctions?.getUiBehavior?.(
                        "verificationSelector",
                    ),
                ).toEqual({
                    "ui:help":
                        "This scheme uses a Checksum algorithm calculation to validate the application image",
                });
                expect(
                    localDefaultModel?.derivedFunctions?.getHelpUrl?.(
                        "verificationSelector",
                        "v2/keyword-lookup?keyword=MELODY.BOOTLOADER8BIT.UI.${verify};&redirect=true",
                    ),
                ).toEqual(
                    "v2/keyword-lookup?keyword=MELODY.BOOTLOADER8BIT.UI.Checksum;&redirect=true",
                );
            });
        });
        describe("Default Return Tests", () => {
            it("Should return the same string when improper component name is given", () => {
                expect(
                    BootloaderDefaultModel?.derivedFunctions?.getHelpUrl?.(
                        "bootEndAddress",
                        "v2/keyword-lookup?keyword=MELODY.BOOTLOADER8BIT.UI.${verify};&redirect=true",
                    ),
                ).toEqual(
                    "v2/keyword-lookup?keyword=MELODY.BOOTLOADER8BIT.UI.${verify};&redirect=true",
                );
            });
        });
    });
});

describe("Tests for import helping functions", () => {
    const testOptions: Processor.Option<any, any, any> = {
        handle: {
            providerId: "module19",
            exportId: "uart_basic",
            label: "module19 : uart_basic",
        },
        name: "UART",
        interfaceId: {
            name: "uart-basic",
            version: "^6.*",
        },
        payload: {
            customName: "UART",
        },
    };
    const testFailureOptions: Processor.Option<any, any, any> = {
        handle: {
            providerId: "module19",
            exportId: "uart_basic",
            label: "module19 : uart_basic",
        },
        name: "UART Driver",
        interfaceId: {
            name: "uart-basic",
            version: "^6.*",
        },
    };

    it("Should succeed with good data", () => {
        expect(
            BootloaderDefaultModel?.derivedFunctions?.importOptionLabel?.(
                "uart_basic",
                testOptions,
            ),
        ).toBe("UART");
    });

    it("Should fail with incorrect or missing data", () => {
        expect(
            BootloaderDefaultModel?.derivedFunctions?.importOptionLabel?.(
                "uart_basic",
                testFailureOptions,
            ),
        ).toBe("UART");
        expect(
            BootloaderDefaultModel?.derivedFunctions?.importName?.("config_request"),
        ).toBe("");
    });
});
