import { CFunction } from "@microchip/scf-interface";
import {
    decimal2Hex,
    parseFunctionAPI,
    parseResetVector,
    replaceAll,
    isPinsDataEmpty,
    isEmptyObject,
    cleanFunctionAPI,
    hasEEPROM,
    parsePIC16ResetVector,
} from "./Helpers";

describe("Helper Tests", () => {
    describe("Hex Conversion Tests", () => {
        it("Should return the converted Hex for any value under 10.", () => {
            expect(decimal2Hex(1)).toBe("0x01");
            expect(decimal2Hex(5)).toBe("0x05");
            expect(decimal2Hex(7)).toBe("0x07");
            expect(decimal2Hex(9)).toBe("0x09");
        });
        it("Should return the converted Hex for any value under 100.", () => {
            expect(decimal2Hex(16)).toBe("0x10");
            expect(decimal2Hex(48)).toBe("0x30");
            expect(decimal2Hex(65)).toBe("0x41");
            expect(decimal2Hex(27)).toBe("0x1B");
        });
        it("Should return the converted Hex for any value under 1000.", () => {
            expect(decimal2Hex(899)).toBe("0x383");
            expect(decimal2Hex(140)).toBe("0x8C");
            expect(decimal2Hex(437)).toBe("0x1B5");
            expect(decimal2Hex(647)).toBe("0x287");
        });
        it("Should return the converted Hex for any value above 1000.", () => {
            expect(decimal2Hex(2234)).toBe("0x8BA");
            expect(decimal2Hex(5896412)).toBe("0x59F8DC");
            expect(decimal2Hex(174348)).toBe("0x2A90C");
            expect(decimal2Hex(21843289874534)).toBe("0x13DDC988A066");
        });
    });

    describe("Parse API Tests", () => {
        const UnimplementedFunctionAPI = "unimplemented";
        const ValidFunctionAPI = {
            name: "Read",
            returns: "uint8_t",
        } as CFunction;
        it("Should return 'unimplemented' when a failure occurs.", () => {
            expect(parseFunctionAPI(UnimplementedFunctionAPI, "FAKENAME")).toBe(
                "unimplemented",
            );
        });

        it("Should return a valid API call.", () => {
            expect(parseFunctionAPI(ValidFunctionAPI, "UART1")).toBe("UART1_Read");
        });
    });
    describe("Parse Reset Vector Address Tests", () => {
        it("Should return the adjusted range if the value is a number.", () => {
            expect(parseResetVector(0x300)).toBe("0-2FF");
            expect(parseResetVector(10)).toBe("0-09");
            expect(parseResetVector(20853)).toBe("0-5174");
        });
        it("Should return the adjusted range if the value is a number for PIC16F.", () => {
            expect(parsePIC16ResetVector(0x300)).toBe("00-2FF");
            expect(parsePIC16ResetVector(10)).toBe("00-09");
            expect(parsePIC16ResetVector(4096)).toBe("00-7FF,800-FFF");
        });
        it("Should fail when incorrect ranges are given.", () => {
            expect(parseResetVector(undefined)).toBe(undefined);
            expect(parseResetVector(0)).toBe(undefined);
        });
    });

    describe("replaceAll function tests", () => {
        it("Should replace all of the target strings with the replace string", () => {
            expect(replaceAll("replacealloftheesinthissentence", "e", "HELLOWORLD")).toBe(
                "rHELLOWORLDplacHELLOWORLDallofthHELLOWORLDHELLOWORLDsinthissHELLOWORLDntHELLOWORLDncHELLOWORLD",
            );

            expect(replaceAll("", "e", "HELLOWORLD")).toBe("");
        });

        describe("isEmptyObject function tests", () => {
            it("Should return a boolean indicating true when any null or missing data is passed. False if the object is 'non-empty'", () => {
                expect(isEmptyObject(undefined)).toBeFalsy();
                expect(isEmptyObject({})).toBeTruthy();
                expect(
                    isEmptyObject({ RF3: { cname: "customName", high: true } }),
                ).toBeFalsy();
            });
        });

        describe("isPinsDataEmpty function tests", () => {
            it("Should return a boolean indicating true when any null or missing data is passed.", () => {
                expect(isPinsDataEmpty("{}")).toBeTruthy();
                expect(isPinsDataEmpty(undefined)).toBeTruthy();
                expect(isPinsDataEmpty({})).toBeTruthy();
                expect(
                    isPinsDataEmpty("{RF3:{cname: 'customName',high: true}}"),
                ).toBeFalsy();
            });
        });

        describe("cleanFunctionAPI function tests", () => {
            it("Should return a boolean indicating true when any null or missing data is passed.", () => {
                expect(cleanFunctionAPI(undefined)).toEqual("unimplemented");
            });

            it("Should return a CFunction type structure", () => {
                expect(
                    cleanFunctionAPI({
                        name: "Read",
                        returns: "uint8_t",
                    } as any),
                ).toEqual({
                    name: "Read",
                    returns: "uint8_t",
                } as CFunction);
            });
        });

        describe("hasEEPROM function tests", () => {
            it("Should return a boolean indicating true when any null or missing data is passed.", () => {
                expect(hasEEPROM(undefined)).toEqual(false);
                expect(hasEEPROM(0)).toEqual(false);
                expect(hasEEPROM(256)).toEqual(true);
            });
        });
    });
});
