import {
    deviceValidator,
    architectureValidator,
    DEVICE_SUPPORT_REG_EXPS,
} from "./deviceValidator";
import { DeviceTestData, Bootloader_Device_Lists } from "./testData";

const test_regularExpression = (
    accept: string[],
    reject: string[],
    REGEXP: RegExp,
): void => {
    for (let index = 0; index < accept.length; index++) {
        it(`${accept[index]} is supported.`, () => {
            expect(REGEXP.test(accept[index])).toBe(true);
        });
    }
    for (let index = 0; index < reject.length; index++) {
        it(`${reject[index]} is NOT supported.`, () => {
            expect(REGEXP.test(reject[index])).toBe(false);
        });
    }
};

describe("Device Validator Tests", () => {
    const NO_SUPPORT: string[] = Array(0).concat(
        Bootloader_Device_Lists.NOSUPPORT.AVRDB,
        Bootloader_Device_Lists.NOSUPPORT.AVRDD,
        Bootloader_Device_Lists.NOSUPPORT.AVRDU,
        Bootloader_Device_Lists.NOSUPPORT.AVREA,
        Bootloader_Device_Lists.NOSUPPORT.Mega,
        Bootloader_Device_Lists.NOSUPPORT.PIC16F153,
        Bootloader_Device_Lists.NOSUPPORT.PIC16F171,
        Bootloader_Device_Lists.NOSUPPORT.PIC16F180,
        Bootloader_Device_Lists.NOSUPPORT.PIC16F181,
        Bootloader_Device_Lists.NOSUPPORT.PIC16F183,
        Bootloader_Device_Lists.NOSUPPORT.PIC16F184,
        Bootloader_Device_Lists.NOSUPPORT.PIC16F191,
        Bootloader_Device_Lists.NOSUPPORT.PIC18FK40,
        Bootloader_Device_Lists.NOSUPPORT.PIC18FK42,
        Bootloader_Device_Lists.NOSUPPORT.PIC18FQ10,
        Bootloader_Device_Lists.NOSUPPORT.PIC18FQ20,
        Bootloader_Device_Lists.NOSUPPORT.PIC18FQ24,
        Bootloader_Device_Lists.NOSUPPORT.PIC18FQ71,
        Bootloader_Device_Lists.NOSUPPORT.PIC18FQ83,
        Bootloader_Device_Lists.NOSUPPORT.PIC18FQ84,
        Bootloader_Device_Lists.NOSUPPORT.Tiny2,
    );
    describe("Intialization Tests", () => {
        it("test that device validator is configureable", () => {
            expect(deviceValidator).toBeDefined();
        });
        it("test that architecure validator is configureable", () => {
            expect(architectureValidator).toBeDefined();
        });
        it("test that architecture validator APIs are reachable", () => {
            expect(architectureValidator.isAVR).toBeDefined();
            expect(architectureValidator.isPIC16).toBeDefined();
            expect(architectureValidator.isPIC18).toBeDefined();
        });
        it("test that device validator APIs are reachable", () => {
            expect(deviceValidator.isSupportedDevice).toBeDefined();
            expect(deviceValidator.isTiny0).toBeDefined();
            expect(deviceValidator.isTiny1).toBeDefined();
            expect(deviceValidator.isPIC18Q).toBeDefined();
            expect(deviceValidator.isAVRXX).toBeDefined();
        });

        it("test that supported device will return undefined on bad input", () => {
            expect(deviceValidator.isSupportedDevice("None")).toBeUndefined();
            expect(deviceValidator.isSupportedDevice("Hello World")).toBeUndefined();
            expect(deviceValidator.isSupportedDevice("")).toBeUndefined();
        });
    });

    describe("test deviceSupport rejection", () => {
        for (let index = 0; index < NO_SUPPORT.length; index++) {
            it(`${NO_SUPPORT[index]} is NOT supported`, () => {
                expect(deviceValidator.isSupportedDevice(NO_SUPPORT[index])).toBeFalsy();
            });
        }
    });

    describe("Device Support Regular Expression Tests", () => {
        let failDevices: string[];
        // AVRDA Test
        describe("AVRDA Tests", () => {
            failDevices = NO_SUPPORT.concat(
                Bootloader_Device_Lists.SUPPORTED.PIC16F152,
                Bootloader_Device_Lists.SUPPORTED.PIC16F188,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ40,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ41,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ43,
                Bootloader_Device_Lists.SUPPORTED.Tiny0,
                Bootloader_Device_Lists.SUPPORTED.Tiny1,
            );

            test_regularExpression(
                Bootloader_Device_Lists.SUPPORTED.AVRDA,
                failDevices,
                DEVICE_SUPPORT_REG_EXPS.AVRDA,
            );
        });

        describe("PIC18 Tests", () => {
            failDevices = NO_SUPPORT.concat(
                Bootloader_Device_Lists.SUPPORTED.PIC16F152,
                Bootloader_Device_Lists.SUPPORTED.PIC16F188,
                Bootloader_Device_Lists.SUPPORTED.AVRDA,
                Bootloader_Device_Lists.SUPPORTED.Tiny0,
                Bootloader_Device_Lists.SUPPORTED.Tiny1,
            );

            test_regularExpression(
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ43.concat(
                    Bootloader_Device_Lists.SUPPORTED.PIC18FQ41,
                    Bootloader_Device_Lists.SUPPORTED.PIC18FQ40,
                ),
                failDevices,
                DEVICE_SUPPORT_REG_EXPS.PIC18,
            );
        });

        describe("PIC16 Tests", () => {
            failDevices = NO_SUPPORT.concat(
                Bootloader_Device_Lists.SUPPORTED.AVRDA,
                Bootloader_Device_Lists.SUPPORTED.Tiny0,
                Bootloader_Device_Lists.SUPPORTED.Tiny1,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ41,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ40,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ43,
            );

            test_regularExpression(
                Bootloader_Device_Lists.SUPPORTED.PIC16F188.concat(
                    Bootloader_Device_Lists.SUPPORTED.PIC16F152,
                ),
                failDevices,
                DEVICE_SUPPORT_REG_EXPS.PIC16,
            );
        });

        describe("ATtiny0 Tests", () => {
            failDevices = NO_SUPPORT.concat(
                Bootloader_Device_Lists.SUPPORTED.AVRDA,
                Bootloader_Device_Lists.SUPPORTED.Tiny1,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ41,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ40,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ43,
                Bootloader_Device_Lists.SUPPORTED.PIC16F152,
                Bootloader_Device_Lists.SUPPORTED.PIC16F188,
            );

            test_regularExpression(
                Bootloader_Device_Lists.SUPPORTED.Tiny0,
                failDevices,
                DEVICE_SUPPORT_REG_EXPS.TINY0,
            );
        });

        describe("ATtiny1 Tests", () => {
            failDevices = NO_SUPPORT.concat(
                Bootloader_Device_Lists.SUPPORTED.AVRDA,
                Bootloader_Device_Lists.SUPPORTED.Tiny0,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ41,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ40,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ43,
                Bootloader_Device_Lists.SUPPORTED.PIC16F152,
                Bootloader_Device_Lists.SUPPORTED.PIC16F188,
            );

            test_regularExpression(
                Bootloader_Device_Lists.SUPPORTED.Tiny1,
                failDevices,
                DEVICE_SUPPORT_REG_EXPS.TINY1,
            );
        });
    });

    describe("AVR Tests", () => {
        it("test that AVR validator is working for all architectures", () => {
            expect(
                architectureValidator.isAVR(DeviceTestData.AVR.architecure),
            ).toBeTruthy();
            expect(
                architectureValidator.isAVR(DeviceTestData.PIC16.architecure),
            ).toBeFalsy();
            expect(
                architectureValidator.isAVR(DeviceTestData.PIC18.architecure),
            ).toBeFalsy();
            expect(architectureValidator.isAVR(undefined)).toBeFalsy();
        });
        describe("Testing Supported Device API with AVR", () => {
            for (
                let index = 0;
                index < DeviceTestData.AVR.supportedDevices.length;
                index++
            ) {
                it(`${DeviceTestData.AVR.supportedDevices[index]} is supported`, () => {
                    expect(
                        deviceValidator.isSupportedDevice(
                            DeviceTestData.AVR.supportedDevices[index],
                        ),
                    ).toBeTruthy();
                });
            }
        });

        describe("Testing Supported Device API with AVR", () => {
            for (
                let index = 0;
                index < DeviceTestData.AVR.supportedDevices.length;
                index++
            ) {
                it(`${DeviceTestData.AVR.supportedDevices[index]} is supported`, () => {
                    expect(
                        deviceValidator.isSupportedDevice(
                            DeviceTestData.AVR.supportedDevices[index],
                        ),
                    ).toBeTruthy();
                });
            }
        });

        // Composite test to check for all supportted devices in the AVR families. (Excluding ATtiny)
        describe("Testing isAVRXX API", () => {
            const AVRXX_DEVICES = Array(0).concat(
                Bootloader_Device_Lists.SUPPORTED.AVRDA,
                // TODO: AVRDB will be supported eventually
            );
            const NOT_AVRXX_DEVICES = NO_SUPPORT.concat(
                Bootloader_Device_Lists.SUPPORTED.PIC16F152,
                Bootloader_Device_Lists.SUPPORTED.PIC16F188,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ40,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ41,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ43,
                Bootloader_Device_Lists.SUPPORTED.Tiny0,
                Bootloader_Device_Lists.SUPPORTED.Tiny1,
            );
            for (let index = 0; index < AVRXX_DEVICES.length; index++) {
                it(`${AVRXX_DEVICES[index]} is supported`, () => {
                    expect(deviceValidator.isAVRXX(AVRXX_DEVICES[index])).toBeTruthy();
                });
            }

            for (let index = 0; index < NOT_AVRXX_DEVICES.length; index++) {
                it(`${NOT_AVRXX_DEVICES[index]} is NOT supported`, () => {
                    expect(deviceValidator.isAVRXX(NOT_AVRXX_DEVICES[index])).toBeFalsy();
                });
            }
        });
    });

    describe("PIC18 Tests", () => {
        it("test that PIC18 validator is working for all architectures", () => {
            expect(
                architectureValidator.isPIC18(DeviceTestData.PIC18.architecure),
            ).toBeTruthy();
            expect(
                architectureValidator.isPIC18(DeviceTestData.AVR.architecure),
            ).toBeFalsy();
            expect(
                architectureValidator.isPIC18(DeviceTestData.PIC16.architecure),
            ).toBeFalsy();
            expect(architectureValidator.isPIC18(undefined)).toBeFalsy();
        });
        describe("Testing Supported Device API with PIC18", () => {
            for (
                let index = 0;
                index < DeviceTestData.PIC18.supportedDevices.length;
                index++
            ) {
                it(`${DeviceTestData.PIC18.supportedDevices[index]} is supported`, () => {
                    expect(
                        deviceValidator.isSupportedDevice(
                            DeviceTestData.PIC18.supportedDevices[index],
                        ),
                    ).toBeTruthy();
                });
            }
        });

        describe("Testing isPIC18Q", () => {
            const PIC18Q_DEVICES = Bootloader_Device_Lists.SUPPORTED.PIC18FQ43.concat(
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ41,
                Bootloader_Device_Lists.SUPPORTED.PIC18FQ40,
            );

            const NOT_PIC18Q_DEVICES = NO_SUPPORT.concat(
                Bootloader_Device_Lists.SUPPORTED.PIC16F152,
                Bootloader_Device_Lists.SUPPORTED.PIC16F188,
                Bootloader_Device_Lists.SUPPORTED.AVRDA,
                Bootloader_Device_Lists.SUPPORTED.Tiny0,
                Bootloader_Device_Lists.SUPPORTED.Tiny1,
            );

            for (let index = 0; index < PIC18Q_DEVICES.length; index++) {
                it(`${PIC18Q_DEVICES[index]} is supported`, () => {
                    expect(deviceValidator.isPIC18Q(PIC18Q_DEVICES[index])).toBeTruthy();
                });
            }

            for (let index = 0; index < NOT_PIC18Q_DEVICES.length; index++) {
                it(`${NOT_PIC18Q_DEVICES[index]} is NOT supported`, () => {
                    expect(
                        deviceValidator.isPIC18Q(NOT_PIC18Q_DEVICES[index]),
                    ).toBeFalsy();
                });
            }
        });
    });

    describe("PIC16 Tests", () => {
        it("test that PIC16 validator is working for all architectures", () => {
            expect(
                architectureValidator.isPIC16(DeviceTestData.PIC16.architecure),
            ).toBeTruthy();
            expect(
                architectureValidator.isPIC16(DeviceTestData.AVR.architecure),
            ).toBeFalsy();
            expect(
                architectureValidator.isPIC16(DeviceTestData.PIC18.architecure),
            ).toBeFalsy();
            expect(architectureValidator.isPIC16(undefined)).toBeFalsy();
        });
        describe("Testing Supported Device API with PIC16", () => {
            for (
                let index = 0;
                index < DeviceTestData.PIC16.supportedDevices.length;
                index++
            ) {
                it(`${DeviceTestData.PIC16.supportedDevices[index]} is supported`, () => {
                    expect(
                        deviceValidator.isSupportedDevice(
                            DeviceTestData.PIC16.supportedDevices[index],
                        ),
                    ).toBeTruthy();
                });
            }
        });

        describe("Testing Supported Device Config Write", () => {
            expect(deviceValidator.supportsConfigWrite("PIC16F15276")).toBeTruthy();
            expect(deviceValidator.supportsConfigWrite("PIC18F57Q43")).toBeTruthy();
            expect(deviceValidator.supportsConfigWrite("PIC18F16Q40")).toBeTruthy();
            expect(deviceValidator.supportsConfigWrite("PIC16F18875")).toBeFalsy();
            expect(deviceValidator.supportsConfigWrite("AVR128DA48")).toBeFalsy();
        });
    });

    describe("ATtiny0 Tests", () => {
        const ACCEPTABLE_TINY0 = Bootloader_Device_Lists.SUPPORTED.Tiny0;
        const UNACCEPTABLE_TINY0 = NO_SUPPORT.concat(
            Bootloader_Device_Lists.SUPPORTED.Tiny1,
            Bootloader_Device_Lists.SUPPORTED.AVRDA,
            Bootloader_Device_Lists.SUPPORTED.PIC16F152,
            Bootloader_Device_Lists.SUPPORTED.PIC18FQ41,
        );

        describe("test that isTiny0 working", () => {
            for (let index = 0; index < UNACCEPTABLE_TINY0.length; index++) {
                it(`${UNACCEPTABLE_TINY0[index]} is NOT supported`, () => {
                    expect(
                        deviceValidator.isTiny0(UNACCEPTABLE_TINY0[index]),
                    ).toBeFalsy();
                });
            }
            for (let index = 0; index < ACCEPTABLE_TINY0.length; index++) {
                it(`${ACCEPTABLE_TINY0[index]} is supported`, () => {
                    expect(deviceValidator.isTiny0(ACCEPTABLE_TINY0[index])).toBeTruthy();
                });
            }
        });
    });

    describe("ATtiny1 Tests", () => {
        const ACCEPTABLE_TINY1 = Bootloader_Device_Lists.SUPPORTED.Tiny1;
        const UNACCEPTABLE_TINY1 = NO_SUPPORT.concat(
            Bootloader_Device_Lists.SUPPORTED.Tiny0,
            Bootloader_Device_Lists.SUPPORTED.AVRDA,
            Bootloader_Device_Lists.SUPPORTED.PIC16F152,
            Bootloader_Device_Lists.SUPPORTED.PIC18FQ41,
        );

        describe("test that isTiny1 working", () => {
            for (let index = 0; index < UNACCEPTABLE_TINY1.length; index++) {
                it(`${UNACCEPTABLE_TINY1[index]} is NOT supported`, () => {
                    expect(
                        deviceValidator.isTiny1(UNACCEPTABLE_TINY1[index]),
                    ).toBeFalsy();
                });
            }
            for (let index = 0; index < ACCEPTABLE_TINY1.length; index++) {
                it(`${ACCEPTABLE_TINY1[index]} is supported`, () => {
                    expect(deviceValidator.isTiny1(ACCEPTABLE_TINY1[index])).toBeTruthy();
                });
            }
        });
    });
});
